from __future__ import annotations

import copy
import functools
import json
import logging
from collections import defaultdict
from typing import Any, DefaultDict, Dict, List, Optional, Set, Text, Tuple, cast

import numpy as np
import structlog
from tqdm import tqdm

import rasa.core.test
import rasa.shared.utils.io
from rasa.core.constants import (
    DEFAULT_CORE_FALLBACK_THRESHOLD,
    POLICY_MAX_HISTORY,
    POLICY_PRIORITY,
    RULE_POLICY_PRIORITY,
)
from rasa.core.featurizers.tracker_featurizers import TrackerFeaturizer
from rasa.core.policies.memoization import MemoizationPolicy
from rasa.core.policies.policy import PolicyPrediction, SupportedData
from rasa.core.training.training import ActionFingerprint, create_action_fingerprints
from rasa.engine.graph import ExecutionContext
from rasa.engine.recipes.default_recipe import DefaultV1Recipe
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.shared.constants import DOCS_URL_RULES
from rasa.shared.core.constants import (
    ACTION_BACK_NAME,
    ACTION_DEFAULT_FALLBACK_NAME,
    ACTION_LISTEN_NAME,
    ACTION_RESTART_NAME,
    ACTION_SESSION_START_NAME,
    ACTIVE_LOOP,
    LOOP_NAME,
    PREVIOUS_ACTION,
    RULE_ONLY_LOOPS,
    RULE_ONLY_SLOTS,
    RULE_SNIPPET_ACTION_NAME,
    SHOULD_NOT_BE_SET,
    SLOTS,
    USER_INTENT_BACK,
    USER_INTENT_RESTART,
    USER_INTENT_SESSION_START,
)
from rasa.shared.core.domain import Domain, InvalidDomain, State
from rasa.shared.core.events import ActionExecuted, LoopInterrupted, UserUttered
from rasa.shared.core.generator import TrackerWithCachedStates
from rasa.shared.core.trackers import (
    DialogueStateTracker,
    get_active_loop_name,
    is_prev_action_listen_in_state,
)
from rasa.shared.exceptions import RasaException
from rasa.shared.nlu.constants import ACTION_NAME, INTENT_NAME_KEY

logger = logging.getLogger(__name__)
structlogger = structlog.get_logger()


# These are Rasa Pro default actions and overrule everything at any time.
DEFAULT_ACTION_MAPPINGS = {
    USER_INTENT_RESTART: ACTION_RESTART_NAME,
    USER_INTENT_BACK: ACTION_BACK_NAME,
    USER_INTENT_SESSION_START: ACTION_SESSION_START_NAME,
}

RULES = "rules"
RULES_FOR_LOOP_UNHAPPY_PATH = "rules_for_loop_unhappy_path"
RULES_NOT_IN_STORIES = "rules_not_in_stories"

LOOP_WAS_INTERRUPTED = "loop_was_interrupted"
DO_NOT_PREDICT_LOOP_ACTION = "do_not_predict_loop_action"

DEFAULT_RULES = "predicting default action with intent "
LOOP_RULES = "handling active loops and forms - "
LOOP_RULES_SEPARATOR = " - "


class InvalidRule(RasaException):
    """Exception that can be raised when rules are not valid."""

    def __init__(self, message: Text) -> None:
        super().__init__(message)
        self.message = message

    def __str__(self) -> Text:
        return self.message + (
            f"\nYou can find more information about the usage of "
            f"rules at {DOCS_URL_RULES}. "
        )


@DefaultV1Recipe.register(
    DefaultV1Recipe.ComponentType.POLICY_WITHOUT_END_TO_END_SUPPORT, is_trainable=True
)
class RulePolicy(MemoizationPolicy):
    """Policy which handles all the rules."""

    # rules use explicit json strings
    ENABLE_FEATURE_STRING_COMPRESSION = False

    # number of user inputs that is allowed in case rules are restricted
    ALLOWED_NUMBER_OF_USER_INPUTS = 1

    @staticmethod
    def supported_data() -> SupportedData:
        """The type of data supported by this policy.

        Returns:
            The data type supported by this policy (ML and rule data).
        """
        return SupportedData.ML_AND_RULE_DATA

    @staticmethod
    def get_default_config() -> Dict[Text, Any]:
        """Returns the default config (see parent class for full docstring)."""
        return {
            # Priority of the policy which is used if multiple policies predict
            # actions with the same confidence.
            POLICY_PRIORITY: RULE_POLICY_PRIORITY,
            # Confidence of the prediction if no rule matched and de-facto
            # threshold for a core fallback.
            "core_fallback_threshold": DEFAULT_CORE_FALLBACK_THRESHOLD,
            # Name of the action which should be predicted if no rule matched.
            "core_fallback_action_name": ACTION_DEFAULT_FALLBACK_NAME,
            # If `True` `core_fallback_action_name` is predicted in case no rule
            # matched.
            "enable_fallback_prediction": True,
            # If `True` rules are restricted to contain a maximum of 1
            # user message. This is used to avoid that users build a state machine
            # using the rules.
            "restrict_rules": True,
            # Whether to check for contradictions between rules and stories
            "check_for_contradictions": True,
            # the policy will use the confidence of NLU on the latest
            # user message to set the confidence of the action
            "use_nlu_confidence_as_score": False,
        }

    def __init__(
        self,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
        featurizer: Optional[TrackerFeaturizer] = None,
        lookup: Optional[Dict] = None,
    ) -> None:
        """Initializes the policy."""
        # max history is set to `None` in order to capture any lengths of rule stories
        config[POLICY_MAX_HISTORY] = None

        super().__init__(
            config, model_storage, resource, execution_context, featurizer, lookup
        )

        self._fallback_action_name = config["core_fallback_action_name"]
        self._enable_fallback_prediction = config["enable_fallback_prediction"]
        self._check_for_contradictions = config["check_for_contradictions"]

        self._rules_sources: DefaultDict[Text, List[Tuple[Text, Text]]] = defaultdict(
            list
        )

    @classmethod
    def raise_if_incompatible_with_domain(
        cls, config: Dict[Text, Any], domain: Domain
    ) -> None:
        """Checks whether the domains action names match the configured fallback.

        Args:
            config: configuration of a `RulePolicy`
            domain: a domain
        Raises:
            `InvalidDomain` if this policy is incompatible with the domain
        """
        fallback_action_name = config.get("core_fallback_action_name", None)
        if (
            fallback_action_name
            and fallback_action_name not in domain.action_names_or_texts
        ):
            raise InvalidDomain(
                f"The fallback action '{fallback_action_name}' which was "
                f"configured for the {RulePolicy.__name__} must be "
                f"present in the domain."
            )

    @staticmethod
    def _is_rule_snippet_state(state: State) -> bool:
        prev_action_name = state.get(PREVIOUS_ACTION, {}).get(ACTION_NAME)
        return prev_action_name == RULE_SNIPPET_ACTION_NAME

    def _create_feature_key(self, states: List[State]) -> Optional[Text]:
        new_states: List[State] = []
        for state in reversed(states):
            if self._is_rule_snippet_state(state):
                # remove all states before RULE_SNIPPET_ACTION_NAME
                break
            new_states.insert(0, state)

        if not new_states:
            return None

        # we sort keys to make sure that the same states
        # represented as dictionaries have the same json strings
        return json.dumps(new_states, sort_keys=True)

    @staticmethod
    def _states_for_unhappy_loop_predictions(states: List[State]) -> List[State]:
        """Modifies the states to create feature keys for loop unhappy path conditions.

        Args:
            states: a representation of a tracker
                as a list of dictionaries containing features

        Returns:
            modified states
        """
        # leave only last 2 dialogue turns to
        # - capture previous meaningful action before action_listen
        # - ignore previous intent
        if len(states) == 1 or not states[-2].get(PREVIOUS_ACTION):
            return [states[-1]]
        else:
            return [{PREVIOUS_ACTION: states[-2][PREVIOUS_ACTION]}, states[-1]]

    @staticmethod
    def _remove_rule_snippet_predictions(lookup: Dict[Text, Text]) -> Dict[Text, Text]:
        # Delete rules if it would predict the RULE_SNIPPET_ACTION_NAME action
        return {
            feature_key: action
            for feature_key, action in lookup.items()
            if action != RULE_SNIPPET_ACTION_NAME
        }

    def _create_loop_unhappy_lookup_from_states(
        self,
        trackers_as_states: List[List[State]],
        trackers_as_actions: List[List[Text]],
    ) -> Dict[Text, Text]:
        """Creates lookup dictionary from the tracker represented as states.

        Args:
            trackers_as_states: representation of the trackers as a list of states
            trackers_as_actions: representation of the trackers as a list of actions

        Returns:
            lookup dictionary
        """
        lookup = {}
        for states, actions in zip(trackers_as_states, trackers_as_actions):
            action = actions[0]
            active_loop = get_active_loop_name(states[-1])
            # even if there are two identical feature keys
            # their loop will be the same
            if not active_loop:
                continue

            states = self._states_for_unhappy_loop_predictions(states)
            feature_key = self._create_feature_key(states)
            if not feature_key:
                continue

            # Since rule snippets and stories inside the loop contain
            # only unhappy paths, notify the loop that
            # it was predicted after an answer to a different question and
            # therefore it should not validate user input
            if (
                # loop is predicted after action_listen in unhappy path,
                # therefore no validation is needed
                is_prev_action_listen_in_state(states[-1]) and action == active_loop
            ):
                lookup[feature_key] = LOOP_WAS_INTERRUPTED
            elif (
                # some action other than active_loop is predicted in unhappy path,
                # therefore active_loop shouldn't be predicted by the rule
                not is_prev_action_listen_in_state(states[-1]) and action != active_loop
            ):
                lookup[feature_key] = DO_NOT_PREDICT_LOOP_ACTION
        return lookup

    def _check_rule_restriction(
        self, rule_trackers: List[TrackerWithCachedStates]
    ) -> None:
        rules_exceeding_max_user_turns = []
        for tracker in rule_trackers:
            number_of_user_uttered = sum(
                isinstance(event, UserUttered) for event in tracker.events
            )
            if number_of_user_uttered > self.ALLOWED_NUMBER_OF_USER_INPUTS:
                rules_exceeding_max_user_turns.append(tracker.sender_id)

        if rules_exceeding_max_user_turns:
            raise InvalidRule(
                f"Found rules '{', '.join(rules_exceeding_max_user_turns)}' "
                f"that contain more than {self.ALLOWED_NUMBER_OF_USER_INPUTS} "
                f"user message. Rules are not meant to hardcode a state machine. "
                f"Please use stories for these cases."
            )

    @staticmethod
    def _expected_but_missing_slots(
        fingerprint: ActionFingerprint, state: State
    ) -> Set[Text]:
        expected_slots = set(fingerprint.slots)
        current_slots = set(state.get(SLOTS, {}).keys())
        # report all slots that are expected but aren't set in current slots
        return expected_slots.difference(current_slots)

    @staticmethod
    def _check_active_loops_fingerprint(
        fingerprint: ActionFingerprint, state: State
    ) -> Set[Optional[Text]]:
        expected_active_loops = set(fingerprint.active_loop)
        # we don't use tracker.active_loop_name
        # because we need to keep should_not_be_set
        current_active_loop = state.get(ACTIVE_LOOP, {}).get(LOOP_NAME)
        if current_active_loop in expected_active_loops:
            # one of expected active loops is set
            return set()

        return expected_active_loops

    @staticmethod
    def _error_messages_from_fingerprints(
        action_name: Text,
        missing_fingerprint_slots: Set[Text],
        fingerprint_active_loops: Set[Text],
        rule_name: Text,
    ) -> List[Text]:
        error_messages = []
        if action_name and missing_fingerprint_slots:
            error_messages.append(
                f"- the action '{action_name}' in rule '{rule_name}' does not set some "
                f"of the slots that it sets in other rules. Slots not set in rule "
                f"'{rule_name}': '{', '.join(missing_fingerprint_slots)}'. Please "
                f"update the rule with an appropriate slot or if it is the last action "
                f"add 'wait_for_user_input: false' after this action."
            )
        if action_name and fingerprint_active_loops:
            # substitute `SHOULD_NOT_BE_SET` with `null` so that users
            # know what to put in their rules
            fingerprint_active_loops = set(
                "null" if active_loop == SHOULD_NOT_BE_SET else active_loop
                for active_loop in fingerprint_active_loops
            )
            # add action_name to active loop so that users
            # know what to put in their rules
            fingerprint_active_loops.add(action_name)

            error_messages.append(
                f"- the form '{action_name}' in rule '{rule_name}' does not set "
                f"the 'active_loop', that it sets in other rules: "
                f"'{', '.join(fingerprint_active_loops)}'. Please update the rule with "
                f"the appropriate 'active loop' property or if it is the last action "
                f"add 'wait_for_user_input: false' after this action."
            )
        return error_messages

    def _check_for_incomplete_rules(
        self, rule_trackers: List[TrackerWithCachedStates], domain: Domain
    ) -> None:
        logger.debug("Started checking if some rules are incomplete.")
        # we need to use only fingerprints from rules
        rule_fingerprints = create_action_fingerprints(rule_trackers, domain)
        if not rule_fingerprints:
            return

        error_messages: List[Text] = []
        for tracker in rule_trackers:
            states = tracker.past_states(domain)
            # the last action is always action listen
            action_names = [
                state.get(PREVIOUS_ACTION, {}).get(ACTION_NAME) for state in states[1:]
            ] + [ACTION_LISTEN_NAME]

            for state, action_name in zip(states, action_names):
                previous_action_name = state.get(PREVIOUS_ACTION, {}).get(ACTION_NAME)
                fingerprint = rule_fingerprints.get(previous_action_name)
                if (
                    not previous_action_name
                    or not fingerprint
                    or action_name == RULE_SNIPPET_ACTION_NAME
                    or previous_action_name == RULE_SNIPPET_ACTION_NAME
                ):
                    # do not check fingerprints for rule snippet action
                    # and don't raise if fingerprints are not satisfied
                    # for a previous action if current action is rule snippet action
                    continue

                missing_expected_slots = self._expected_but_missing_slots(
                    fingerprint, state
                )
                expected_active_loops = self._check_active_loops_fingerprint(
                    fingerprint, state
                )
                error_messages.extend(
                    self._error_messages_from_fingerprints(
                        previous_action_name,
                        missing_expected_slots,
                        expected_active_loops,
                        tracker.sender_id,
                    )
                )

        if error_messages:
            error_text = "\n".join(error_messages)
            raise InvalidRule(
                f"\nIncomplete rules found🚨\n\n{error_text}\n"
                f"Please note that if some slots or active loops should not be set "
                f"during prediction you need to explicitly set them to 'null' in the "
                f"rules."
            )

        logger.debug("Found no incompletions in rules.")

    @staticmethod
    def _get_slots_loops_from_states(
        trackers_as_states: List[List[State]],
    ) -> Tuple[Set[Text], Set[Text]]:
        slots = set()
        loops = set()
        for states in trackers_as_states:
            for state in states:
                slots.update(set(state.get(SLOTS, {}).keys()))
                # FIXME: ideally we have better annotation for State, TypedDict
                # could work but support in mypy is very limited. Dataclass are
                # another option
                active_loop = cast(Text, state.get(ACTIVE_LOOP, {}).get(LOOP_NAME))
                if active_loop:
                    loops.add(active_loop)
        return slots, loops

    def _find_rule_only_slots_loops(
        self,
        rule_trackers_as_states: List[List[State]],
        story_trackers_as_states: List[List[State]],
    ) -> Tuple[List[Text], List[Text]]:
        rule_slots, rule_loops = self._get_slots_loops_from_states(
            rule_trackers_as_states
        )
        story_slots, story_loops = self._get_slots_loops_from_states(
            story_trackers_as_states
        )

        # set is not json serializable, so convert to list
        return (
            list(rule_slots - story_slots - {SHOULD_NOT_BE_SET}),
            list(rule_loops - story_loops - {SHOULD_NOT_BE_SET}),
        )

    def _predict_next_action(
        self, tracker: TrackerWithCachedStates, domain: Domain
    ) -> Tuple[Optional[Text], Optional[Text]]:
        prediction, prediction_source = self._predict(tracker, domain)
        probabilities = prediction.probabilities
        # do not raise an error if RulePolicy didn't predict anything for stories;
        # however for rules RulePolicy should always predict an action
        predicted_action_name = None
        if (
            probabilities != self._default_predictions(domain)
            or tracker.is_rule_tracker
        ):
            predicted_action_name = domain.action_names_or_texts[
                np.argmax(probabilities)
            ]

        return predicted_action_name, prediction_source

    def _predicted_action_name(
        self, tracker: TrackerWithCachedStates, domain: Domain, gold_action_name: Text
    ) -> Tuple[Optional[Text], Optional[Text]]:
        predicted_action_name, prediction_source = self._predict_next_action(
            tracker, domain
        )
        # if there is an active_loop,
        # RulePolicy will always predict active_loop first,
        # but inside loop unhappy path there might be another action
        if (
            tracker.active_loop_name
            and predicted_action_name != gold_action_name
            and predicted_action_name == tracker.active_loop_name
        ):
            rasa.core.test.emulate_loop_rejection(tracker)
            predicted_action_name, prediction_source = self._predict_next_action(
                tracker, domain
            )

        return predicted_action_name, prediction_source

    def _collect_sources(
        self,
        tracker: TrackerWithCachedStates,
        predicted_action_name: Optional[Text],
        gold_action_name: Optional[Text],
        prediction_source: Text,
    ) -> None:
        # we need to remember which action should be predicted by the rule
        # in order to correctly output the names of the contradicting rules
        rule_name = tracker.sender_id

        if prediction_source is not None and (
            prediction_source.startswith(DEFAULT_RULES)
            or prediction_source.startswith(LOOP_RULES)
        ):
            # the real gold action contradict the one in the rules in this case
            gold_action_name = predicted_action_name
            rule_name = prediction_source

        self._rules_sources[prediction_source].append((rule_name, gold_action_name))

    @staticmethod
    def _default_sources() -> Set[Text]:
        return {
            DEFAULT_RULES + default_intent
            for default_intent in DEFAULT_ACTION_MAPPINGS.keys()
        }

    @staticmethod
    def _handling_loop_sources(domain: Domain) -> Set[Text]:
        loop_sources = set()
        for loop_name in domain.form_names:
            loop_sources.add(LOOP_RULES + loop_name)
            loop_sources.add(
                LOOP_RULES + loop_name + LOOP_RULES_SEPARATOR + ACTION_LISTEN_NAME
            )
        return loop_sources

    def _should_delete(
        self,
        prediction_source: Text,
        tracker: TrackerWithCachedStates,
        predicted_action_name: Text,
    ) -> bool:
        """Checks whether this contradiction is due to action, intent pair.

        Args:
            prediction_source: the states that result in the prediction
            tracker: the tracker that raises the contradiction
            predicted_action_name: the action that was predicted

        Returns:
            true if the contradiction is a result of an action, intent pair in the rule.
        """
        if (
            # only apply to contradicting story, not rule
            tracker.is_rule_tracker
            # only apply for prediction after unpredictable action
            or prediction_source.count(PREVIOUS_ACTION) > 1
            # only apply for prediction of action_listen
            or predicted_action_name != ACTION_LISTEN_NAME
        ):
            return False
        for source in self.lookup[RULES]:
            # remove rule only if another action is predicted after action_listen
            if (
                source.startswith(prediction_source[:-2])
                and not prediction_source == source
            ):
                return True
        return False

    def _check_prediction(
        self,
        tracker: TrackerWithCachedStates,
        predicted_action_name: Optional[Text],
        gold_action_name: Text,
        prediction_source: Optional[Text],
    ) -> List[Text]:
        # FIXME: `predicted_action_name` and `prediction_source` are
        # either None together or defined together. This could be improved
        # by better typing in this class, but requires some refactoring
        if (
            not predicted_action_name
            or not prediction_source
            or predicted_action_name == gold_action_name
        ):
            return []

        if self._should_delete(prediction_source, tracker, predicted_action_name):
            self.lookup[RULES].pop(prediction_source)
            return []

        tracker_type = "rule" if tracker.is_rule_tracker else "story"
        contradicting_rules = {
            rule_name
            for rule_name, action_name in self._rules_sources[prediction_source]
            if action_name != gold_action_name
        }

        if not contradicting_rules:
            return []

        error_message = (
            f"- the prediction of the action '{gold_action_name}' in {tracker_type} "
            f"'{tracker.sender_id}' "
            f"is contradicting with rule(s) '{', '.join(contradicting_rules)}'"
        )
        # outputting predicted action 'action_default_fallback' is confusing
        if predicted_action_name != self._fallback_action_name:
            error_message += f" which predicted action '{predicted_action_name}'"

        return [error_message + "."]

    def _run_prediction_on_trackers(
        self,
        trackers: List[TrackerWithCachedStates],
        domain: Domain,
        collect_sources: bool,
    ) -> Tuple[List[Text], Set[Optional[Text]]]:
        if collect_sources:
            self._rules_sources = defaultdict(list)

        error_messages = []
        rules_used_in_stories = set()
        pbar = tqdm(
            trackers,
            desc="Processed trackers",
            disable=rasa.shared.utils.io.is_logging_disabled(),
        )
        for tracker in pbar:
            running_tracker = tracker.init_copy()
            running_tracker.sender_id = tracker.sender_id
            # the first action is always unpredictable
            next_action_is_unpredictable = True
            for event in tracker.applied_events(True):
                if not isinstance(event, ActionExecuted):
                    running_tracker.update(event)
                    continue

                if event.action_name == RULE_SNIPPET_ACTION_NAME:
                    # notify that the action after RULE_SNIPPET_ACTION_NAME is
                    # unpredictable
                    next_action_is_unpredictable = True
                    running_tracker.update(event)
                    continue

                # do not run prediction on unpredictable actions
                if next_action_is_unpredictable or event.unpredictable:
                    next_action_is_unpredictable = False  # reset unpredictability
                    running_tracker.update(event)
                    continue

                gold_action_name = event.action_name or event.action_text
                predicted_action_name, prediction_source = self._predicted_action_name(
                    running_tracker, domain, gold_action_name
                )
                if collect_sources:
                    if prediction_source:
                        self._collect_sources(
                            running_tracker,
                            predicted_action_name,
                            gold_action_name,
                            prediction_source,
                        )
                else:
                    # to be able to remove only rules turns from the dialogue history
                    # for ML policies,
                    # we need to know which rules were used in ML trackers
                    if (
                        not tracker.is_rule_tracker
                        and predicted_action_name == gold_action_name
                    ):
                        rules_used_in_stories.add(prediction_source)

                    error_messages += self._check_prediction(
                        running_tracker,
                        predicted_action_name,
                        gold_action_name,
                        prediction_source,
                    )

                running_tracker.update(event)

        return error_messages, rules_used_in_stories

    def _collect_rule_sources(
        self, rule_trackers: List[TrackerWithCachedStates], domain: Domain
    ) -> None:
        self._run_prediction_on_trackers(rule_trackers, domain, collect_sources=True)

    def _find_contradicting_and_used_in_stories_rules(
        self, trackers: List[TrackerWithCachedStates], domain: Domain
    ) -> Tuple[List[Text], Set[Optional[Text]]]:
        return self._run_prediction_on_trackers(trackers, domain, collect_sources=False)

    def _analyze_rules(
        self,
        rule_trackers: List[TrackerWithCachedStates],
        all_trackers: List[TrackerWithCachedStates],
        domain: Domain,
    ) -> List[Text]:
        """Analyzes learned rules by running prediction on training trackers.

        This method collects error messages for contradicting rules
        and creates the lookup for rules that are not present in the stories.

        Args:
            rule_trackers: The list of the rule trackers.
            all_trackers: The list of all trackers.
            domain: The domain.

        Returns:
             Rules that are not present in the stories.
        """
        logger.debug("Started checking rules and stories for contradictions.")
        # during training we run `predict_action_probabilities` to check for
        # contradicting rules.
        # We silent prediction debug to avoid too many logs during these checks.
        logger_level = logger.level
        logger.setLevel(logging.WARNING)

        # we need to run prediction on rule trackers twice, because we need to collect
        # the information about which rule snippets contributed to the learned rules
        self._collect_rule_sources(rule_trackers, domain)
        (
            error_messages,
            rules_used_in_stories,
        ) = self._find_contradicting_and_used_in_stories_rules(all_trackers, domain)

        logger.setLevel(logger_level)  # reset logger level
        if error_messages:
            error_text = "\n".join(error_messages)
            raise InvalidRule(
                f"\nContradicting rules or stories found 🚨\n\n{error_text}\n"
                f"Please update your stories and rules so that they don't contradict "
                f"each other."
            )

        logger.debug("Found no contradicting rules.")
        all_rules = (
            set(self._rules_sources.keys())
            | self._default_sources()
            | self._handling_loop_sources(domain)
        )
        # set is not json serializable, so convert to list
        return list(all_rules - rules_used_in_stories)

    def _create_lookup_from_trackers(
        self,
        rule_trackers: List[TrackerWithCachedStates],
        story_trackers: List[TrackerWithCachedStates],
        domain: Domain,
    ) -> None:
        (
            rule_trackers_as_states,
            rule_trackers_as_actions,
        ) = self.featurizer.training_states_and_labels(
            rule_trackers, domain, omit_unset_slots=True
        )

        rules_lookup = self._create_lookup_from_states(
            rule_trackers_as_states, rule_trackers_as_actions
        )
        self.lookup[RULES] = self._remove_rule_snippet_predictions(rules_lookup)

        (
            story_trackers_as_states,
            story_trackers_as_actions,
        ) = self.featurizer.training_states_and_labels(story_trackers, domain)

        if self._check_for_contradictions:
            (
                self.lookup[RULE_ONLY_SLOTS],
                self.lookup[RULE_ONLY_LOOPS],
            ) = self._find_rule_only_slots_loops(
                rule_trackers_as_states, story_trackers_as_states
            )

        # use all trackers to find negative rules in unhappy paths
        trackers_as_states = rule_trackers_as_states + story_trackers_as_states
        trackers_as_actions = rule_trackers_as_actions + story_trackers_as_actions

        # negative rules are not anti-rules, they are auxiliary to actual rules
        self.lookup[RULES_FOR_LOOP_UNHAPPY_PATH] = (
            self._create_loop_unhappy_lookup_from_states(
                trackers_as_states, trackers_as_actions
            )
        )

    def train(
        self,
        training_trackers: List[TrackerWithCachedStates],
        domain: Domain,
        **kwargs: Any,
    ) -> Resource:
        """Trains the policy on given training trackers.

        Args:
            training_trackers: The list of the trackers.
            domain: The domain.
            **kwargs: Additional arguments.

        Returns:
            The resource which can be used to load the trained policy.
        """
        self.raise_if_incompatible_with_domain(self.config, domain)

        # only consider original trackers (no augmented ones)
        training_trackers = [
            t for t in training_trackers if not getattr(t, "is_augmented", False)
        ]
        # trackers from rule-based training data
        rule_trackers = [t for t in training_trackers if t.is_rule_tracker]
        if self.config["restrict_rules"]:
            self._check_rule_restriction(rule_trackers)
        if self._check_for_contradictions:
            self._check_for_incomplete_rules(rule_trackers, domain)

        # trackers from ML-based training data
        story_trackers = [t for t in training_trackers if not t.is_rule_tracker]

        self._create_lookup_from_trackers(rule_trackers, story_trackers, domain)

        # make this configurable because checking might take a lot of time
        if self._check_for_contradictions:
            # using trackers here might not be the most efficient way, however
            # it allows us to directly test `predict_action_probabilities` method
            self.lookup[RULES_NOT_IN_STORIES] = self._analyze_rules(
                rule_trackers, training_trackers, domain
            )

        logger.debug(f"Memorized '{len(self.lookup[RULES])}' unique rules.")

        self.persist()

        return self._resource

    @staticmethod
    def _does_rule_match_state(rule_state: State, conversation_state: State) -> bool:
        for state_type, rule_sub_state in rule_state.items():
            conversation_sub_state = conversation_state.get(state_type, {})
            for key, value_from_rules in rule_sub_state.items():
                if isinstance(value_from_rules, list):
                    # json dumps and loads tuples as lists,
                    # so we need to convert them back
                    value_from_rules = tuple(value_from_rules)
                value_from_conversation = conversation_sub_state.get(key)
                if (
                    # value should be set, therefore
                    # check whether it is the same as in the state
                    value_from_rules
                    and value_from_rules != SHOULD_NOT_BE_SET
                    and value_from_conversation != value_from_rules
                ) or (
                    # value shouldn't be set, therefore
                    # it should be None or non existent in the state
                    value_from_rules == SHOULD_NOT_BE_SET
                    and value_from_conversation
                    # during training `SHOULD_NOT_BE_SET` is provided. Hence, we also
                    # have to check for the value of the slot state
                    and value_from_conversation != SHOULD_NOT_BE_SET
                ):
                    return False

        return True

    @staticmethod
    # This function is called a lot (e.g. for checking contradictions) so we cache
    # its results.
    @functools.lru_cache(maxsize=1000)
    def _rule_key_to_state(rule_key: Text) -> List[State]:
        return json.loads(rule_key)

    def _is_rule_applicable(
        self, rule_key: Text, turn_index: int, conversation_state: State
    ) -> bool:
        """Checks if rule is satisfied with current state at turn.

        Args:
            rule_key: the textual representation of learned rule
            turn_index: index of a current dialogue turn
            conversation_state: the state that corresponds to turn_index

        Returns:
            a boolean that says whether the rule is applicable to current state
        """
        # turn_index goes back in time
        reversed_rule_states = list(reversed(self._rule_key_to_state(rule_key)))

        # the rule must be applicable because we got (without any applicability issues)
        # further in the conversation history than the rule's length
        if turn_index >= len(reversed_rule_states):
            return True

        # a state has previous action if and only if it is not a conversation start
        # state
        current_previous_action = conversation_state.get(PREVIOUS_ACTION)
        rule_previous_action = reversed_rule_states[turn_index].get(PREVIOUS_ACTION)

        # current conversation state and rule state are conversation starters.
        # any slots with initial_value set will necessarily be in both states and don't
        # need to be checked.
        if not rule_previous_action and not current_previous_action:
            return True

        # current rule state is a conversation starter (due to conversation_start: true)
        # but current conversation state is not.
        # or
        # current conversation state is a starter
        # but current rule state is not.
        if not rule_previous_action or not current_previous_action:
            return False

        # check: current rule state features are present in current conversation state
        return self._does_rule_match_state(
            reversed_rule_states[turn_index], conversation_state
        )

    def _get_possible_keys(
        self, lookup: Dict[Text, Text], states: List[State]
    ) -> Set[Text]:
        possible_keys = set(lookup.keys())
        for i, state in enumerate(reversed(states)):
            # find rule keys that correspond to current state
            possible_keys = set(
                filter(
                    lambda _key: self._is_rule_applicable(_key, i, state), possible_keys
                )
            )
        return possible_keys

    @staticmethod
    def _find_action_from_default_actions(
        tracker: DialogueStateTracker,
    ) -> Tuple[Optional[Text], Optional[Text]]:
        if (
            not tracker.latest_action_name == ACTION_LISTEN_NAME
            or not tracker.latest_message
        ):
            return None, None

        intent_name = tracker.latest_message.intent.get(INTENT_NAME_KEY)
        if intent_name is None:
            return None, None

        default_action_name = DEFAULT_ACTION_MAPPINGS.get(intent_name)
        if default_action_name is None:
            return None, None

        logger.debug(f"Predicted default action '{default_action_name}'.")
        return (
            default_action_name,
            # create prediction source that corresponds to one of
            # default prediction sources in `_default_sources()`
            DEFAULT_RULES + intent_name,
        )

    @staticmethod
    def _find_action_from_loop_happy_path(
        tracker: DialogueStateTracker,
    ) -> Tuple[Optional[Text], Optional[Text]]:
        active_loop_name = tracker.active_loop_name
        if active_loop_name is None:
            return None, None

        active_loop_rejected = tracker.is_active_loop_rejected
        should_predict_loop = (
            not active_loop_rejected
            and tracker.latest_action
            and tracker.latest_action.get(ACTION_NAME) != active_loop_name
        )
        should_predict_listen = (
            not active_loop_rejected and tracker.latest_action_name == active_loop_name
        )

        if should_predict_loop:
            logger.debug(f"Predicted loop '{active_loop_name}'.")
            return active_loop_name, LOOP_RULES + active_loop_name

        # predict `action_listen` if loop action was run successfully
        if should_predict_listen:
            logger.debug(
                f"Predicted '{ACTION_LISTEN_NAME}' after loop '{active_loop_name}'."
            )
            return (
                ACTION_LISTEN_NAME,
                (
                    f"{LOOP_RULES}{active_loop_name}"
                    f"{LOOP_RULES_SEPARATOR}{ACTION_LISTEN_NAME}"
                ),
            )

        return None, None

    def _find_action_from_rules(
        self,
        tracker: DialogueStateTracker,
        domain: Domain,
        use_text_for_last_user_input: bool,
    ) -> Tuple[Optional[Text], Optional[Text], bool]:
        """Predicts the next action based on the memoized rules.

        Args:
            tracker: The current conversation tracker.
            domain: The domain of the current model.
            use_text_for_last_user_input: `True` if text of last user message
                should be used for the prediction. `False` if intent should be used.

        Returns:
            A tuple of the predicted action name or text (or `None` if no matching rule
            was found), a description of the matching rule, and `True` if a loop action
            was predicted after the loop has been in an unhappy path before.
        """
        if (
            use_text_for_last_user_input
            and not tracker.latest_action_name == ACTION_LISTEN_NAME
        ):
            # make text prediction only directly after user utterance
            # because we've otherwise already decided whether to use
            # the text or the intent
            return None, None, False

        states = self._prediction_states(
            tracker,
            domain,
            use_text_for_last_user_input,
            rule_only_data=self._get_rule_only_data(),
        )

        current_states = self.format_tracker_states(states)
        structlogger.debug(
            "rule_policy.actions.find", current_states=copy.deepcopy(current_states)
        )

        # Tracks if we are returning after an unhappy loop path. If this becomes `True`
        # the policy returns an event which notifies the loop action that it
        # is returning after an unhappy path. For example, the `FormAction` uses this
        # to skip the validation of slots for its first execution after an unhappy path.
        returning_from_unhappy_path = False

        rule_keys = self._get_possible_keys(self.lookup[RULES], states)
        predicted_action_name = None
        best_rule_key = ""
        if rule_keys:
            # if there are several rules,
            # it should mean that some rule is a subset of another rule
            # therefore we pick a rule of maximum length
            best_rule_key = max(rule_keys, key=len)
            predicted_action_name = self.lookup[RULES].get(best_rule_key)

        active_loop_name = tracker.active_loop_name
        if active_loop_name:
            # find rules for unhappy path of the loop
            loop_unhappy_keys = self._get_possible_keys(
                self.lookup[RULES_FOR_LOOP_UNHAPPY_PATH], states
            )
            # there could be several unhappy path conditions
            unhappy_path_conditions = [
                self.lookup[RULES_FOR_LOOP_UNHAPPY_PATH].get(key)
                for key in loop_unhappy_keys
            ]

            # Check if a rule that predicted action_listen
            # was applied inside the loop.
            # Rules might not explicitly switch back to the loop.
            # Hence, we have to take care of that.
            predicted_listen_from_general_rule = (
                predicted_action_name == ACTION_LISTEN_NAME
                and not get_active_loop_name(self._rule_key_to_state(best_rule_key)[-1])
            )
            if predicted_listen_from_general_rule:
                if DO_NOT_PREDICT_LOOP_ACTION not in unhappy_path_conditions:
                    # negative rules don't contain a key that corresponds to
                    # the fact that active_loop shouldn't be predicted
                    logger.debug(
                        f"Predicted loop '{active_loop_name}' by overwriting "
                        f"'{ACTION_LISTEN_NAME}' predicted by general rule."
                    )
                    return (
                        active_loop_name,
                        best_rule_key,
                        returning_from_unhappy_path,
                    )

                # do not predict anything
                predicted_action_name = None

            if LOOP_WAS_INTERRUPTED in unhappy_path_conditions:
                logger.debug(
                    "Returning from unhappy path. Loop will be notified that "
                    "it was interrupted."
                )
                returning_from_unhappy_path = True

        if predicted_action_name is not None:
            logger.debug(
                f"There is a rule for the next action '{predicted_action_name}'."
            )
        else:
            logger.debug("There is no applicable rule.")

        # if we didn't predict anything from the rules, then the feature key created
        # from states can be used as an indicator that this state will lead to fallback
        return (
            predicted_action_name,
            best_rule_key or self._create_feature_key(states),
            returning_from_unhappy_path,
        )

    async def predict_action_probabilities(
        self,
        tracker: DialogueStateTracker,
        domain: Domain,
        rule_only_data: Optional[Dict[Text, Any]] = None,
        **kwargs: Any,
    ) -> PolicyPrediction:
        """Predicts the next action (see parent class for more information)."""
        if self.should_abstain_in_coexistence(tracker, False):
            # don't use self._default_predictions as this might have a different
            # probability for the fallback action, and we want to have all probabilities
            # set to 0.0
            return self._prediction(super()._default_predictions(domain))

        prediction, _ = self._predict(tracker, domain)
        return prediction

    def _predict(
        self, tracker: DialogueStateTracker, domain: Domain
    ) -> Tuple[PolicyPrediction, Optional[Text]]:
        (
            rules_action_name_from_text,
            prediction_source_from_text,
            returning_from_unhappy_path_from_text,
        ) = self._find_action_from_rules(
            tracker, domain, use_text_for_last_user_input=True
        )

        # Rasa Pro default actions overrule anything. If users want to achieve
        # the same, they need to write a rule or make sure that their loop rejects
        # accordingly.
        (
            default_action_name,
            default_prediction_source,
        ) = self._find_action_from_default_actions(tracker)

        # text has priority over intents including default,
        # however loop happy path has priority over rules prediction
        if default_action_name and not rules_action_name_from_text:
            return (
                self._rule_prediction(
                    self._prediction_result(default_action_name, tracker, domain),
                    default_prediction_source,
                ),
                default_prediction_source,
            )

        # A loop has priority over any other rule except defaults.
        # The rules or any other prediction will be applied only if a loop was rejected.
        # If we are in a loop, and the loop didn't run previously or rejected, we can
        # simply force predict the loop.
        (
            loop_happy_path_action_name,
            loop_happy_path_prediction_source,
        ) = self._find_action_from_loop_happy_path(tracker)
        if loop_happy_path_action_name:
            # this prediction doesn't use user input
            # and happy user input anyhow should be ignored during featurization
            return (
                self._rule_prediction(
                    self._prediction_result(
                        loop_happy_path_action_name, tracker, domain
                    ),
                    loop_happy_path_prediction_source,
                    is_no_user_prediction=True,
                ),
                loop_happy_path_prediction_source,
            )

        # predict rules from text first
        if rules_action_name_from_text:
            return (
                self._rule_prediction(
                    self._prediction_result(
                        rules_action_name_from_text, tracker, domain
                    ),
                    prediction_source_from_text,
                    returning_from_unhappy_path=returning_from_unhappy_path_from_text,
                    is_end_to_end_prediction=True,
                ),
                prediction_source_from_text,
            )

        (
            rules_action_name_from_intent,
            # we want to remember the source even if rules didn't predict any action
            prediction_source_from_intent,
            returning_from_unhappy_path_from_intent,
        ) = self._find_action_from_rules(
            tracker, domain, use_text_for_last_user_input=False
        )
        if rules_action_name_from_intent:
            probabilities = self._prediction_result(
                rules_action_name_from_intent, tracker, domain
            )
        else:
            probabilities = self._default_predictions(domain)

        return (
            self._rule_prediction(
                probabilities,
                prediction_source_from_intent,
                returning_from_unhappy_path=(
                    # returning_from_unhappy_path is a negative condition,
                    # so `or` should be applied
                    returning_from_unhappy_path_from_text
                    or returning_from_unhappy_path_from_intent
                ),
                is_end_to_end_prediction=False,
            ),
            prediction_source_from_intent,
        )

    def _rule_prediction(
        self,
        probabilities: List[float],
        prediction_source: Text,
        returning_from_unhappy_path: bool = False,
        is_end_to_end_prediction: bool = False,
        is_no_user_prediction: bool = False,
    ) -> PolicyPrediction:
        return PolicyPrediction(
            probabilities,
            self.__class__.__name__,
            self.priority,
            events=[LoopInterrupted(True)] if returning_from_unhappy_path else [],
            is_end_to_end_prediction=is_end_to_end_prediction,
            is_no_user_prediction=is_no_user_prediction,
            hide_rule_turn=(
                True
                if prediction_source in self.lookup.get(RULES_NOT_IN_STORIES, [])
                else False
            ),
        )

    def _default_predictions(self, domain: Domain) -> List[float]:
        result = super()._default_predictions(domain)

        if self._enable_fallback_prediction:
            result[domain.index_for_action(self._fallback_action_name)] = self.config[
                "core_fallback_threshold"
            ]

        return result

    def persist(self) -> None:
        """Persists trained `RulePolicy`."""
        super().persist()
        with self._model_storage.write_to(self._resource) as directory:
            rule_only_data = self._get_rule_only_data()
            rasa.shared.utils.io.dump_obj_as_json_to_file(
                directory / "rule_only_data.json", rule_only_data
            )

    def _metadata(self) -> Dict[Text, Any]:
        return {"lookup": self.lookup}

    @classmethod
    def _metadata_filename(cls) -> Text:
        return "rule_policy.json"

    def _get_rule_only_data(self) -> Dict[Text, Any]:
        """Gets the slots and loops that are used only in rule data.

        Returns:
            Slots and loops that are used only in rule data.
        """
        return {
            key: self.lookup.get(key, []) for key in [RULE_ONLY_SLOTS, RULE_ONLY_LOOPS]
        }



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Optional

from rasa.dialogue_understanding.stack.frames.dialogue_stack_frame import (
    DialogueStackFrame,
)
from rasa.shared.core.flows import Flow, FlowsList, FlowStep
from rasa.shared.core.flows.steps.constants import START_STEP
from rasa.shared.exceptions import RasaException


class InvalidFlowStackFrameType(RasaException):
    """Raised if the stack frame type is invalid."""

    def __init__(self, frame_type: Optional[str]) -> None:
        """Creates a `InvalidFlowStackFrameType`.

        Args:
            frame_type: The invalid stack frame type.
        """
        super().__init__(f"Invalid stack frame type '{frame_type}'.")


class FlowStackFrameType(str, Enum):
    INTERRUPT = "interrupt"
    """The frame is an interrupt frame.

    This means that the previous flow was interrupted by this flow. An
    interrupt should be used for frames that span multiple turns and
    where we expect the user needing help to get back to the previous
    flow."""
    LINK = "link"
    """The frame is a link frame.

    This means that the previous flow linked to this flow."""

    CALL = "call"
    """The frame is a flow that is called from another flow."""

    REGULAR = "regular"
    """The frame is a regular frame.

    In all other cases, this is the case."""

    @staticmethod
    def from_str(typ: Optional[str]) -> "FlowStackFrameType":
        """Creates a `FlowStackFrameType` from a string.

        Args:
            typ: The string to create the `FlowStackFrameType` from.

        Returns:
        The created `FlowStackFrameType`.
        """
        if typ is None:
            return FlowStackFrameType.REGULAR
        elif typ == FlowStackFrameType.INTERRUPT.value:
            return FlowStackFrameType.INTERRUPT
        elif typ == FlowStackFrameType.CALL.value:
            return FlowStackFrameType.CALL
        elif typ == FlowStackFrameType.LINK.value:
            return FlowStackFrameType.LINK
        elif typ == FlowStackFrameType.REGULAR.value:
            return FlowStackFrameType.REGULAR
        else:
            raise InvalidFlowStackFrameType(typ)


class InvalidFlowIdException(Exception):
    """Raised if the flow ID is invalid."""

    def __init__(self, flow_id: str) -> None:
        """Creates a `InvalidFlowIdException`.

        Args:
            flow_id: The invalid flow ID.
        """
        super().__init__(f"Invalid flow ID '{flow_id}'.")


class InvalidFlowStepIdException(Exception):
    """Raised if the flow step ID is invalid."""

    def __init__(self, flow_id: str, step_id: str) -> None:
        """Creates a `InvalidFlowStepIdException`.

        Args:
            flow_id: The invalid flow ID.
            step_id: The invalid flow step ID.
        """
        super().__init__(f"Invalid flow step ID '{step_id}' for flow '{flow_id}'.")


class InvalidAgentState(RasaException):
    """Raised if the agent state is invalid."""

    def __init__(self, invalid_state: str) -> None:
        """Creates a `InvalidAgentState`.

        Args:
            invalid_state: The invalid agent state.
        """
        super().__init__(f"Invalid agent state '{invalid_state}'.")


@dataclass
class BaseFlowStackFrame(DialogueStackFrame):
    flow_id: str = ""  # needed to avoid "default arg before non-default" error
    """The ID of the current flow."""
    step_id: str = START_STEP
    """The ID of the current step."""

    def flow(self, all_flows: FlowsList) -> Flow:
        """Returns the current flow.

        Args:
            all_flows: All flows in the assistant.

        Returns:
        The current flow.
        """
        flow = all_flows.flow_by_id(self.flow_id)
        if not flow:
            # we shouldn't ever end up with a frame that belongs to a non
            # existing flow, but if we do, we should raise an error
            raise InvalidFlowIdException(self.flow_id)
        return flow

    def step(self, all_flows: FlowsList) -> FlowStep:
        """Returns the current flow step.

        Args:
            all_flows: All flows in the assistant.

        Returns:
            The current flow step.
        """
        flow = self.flow(all_flows)

        step_id = self.step_id
        # in 3.11.4 we added the flow_id as a prefix to the step_id
        # this causes issues when loading old dialogues as the prefix is missing
        # (see https://rasahq.atlassian.net/jira/software/c/projects/ENG/boards/43?selectedIssue=ENG-1939)
        # so we try to find the step by adding the flow prefix to old step_ids as well
        # TODO: remove this in 4.0.0
        alternative_step_id = f"{self.flow_id}_{self.step_id}"

        step = flow.step_by_id(step_id) or flow.step_by_id(alternative_step_id)

        if not step:
            # we shouldn't ever end up with a frame that belongs to a non
            # existing step, but if we do, we should raise an error
            raise InvalidFlowStepIdException(self.flow_id, self.step_id)
        return step


@dataclass
class UserFlowStackFrame(BaseFlowStackFrame):
    frame_type: FlowStackFrameType = FlowStackFrameType.REGULAR
    """The type of the frame. Defaults to `StackFrameType.REGULAR`."""
    suspended_agent_frames: Optional[List[Dict[str, Any]]] = None
    """Agent stack frames (as dicts) to restore when reaching that agent's step.

    Used when rewinding the stack (e.g. slot correction or agent restart) so
    interrupted agents can be resumed. Old trackers may have no key; treat as None.
    """

    @classmethod
    def type(cls) -> str:
        """Returns the type of the frame."""
        return "flow"

    def as_dict(self) -> Dict[str, Any]:
        """Serialize to dict; omit suspended_agent_frames when None to avoid bloat."""
        data = super().as_dict()
        if data.get("suspended_agent_frames") is None:
            data.pop("suspended_agent_frames", None)
        return data

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> UserFlowStackFrame:
        """Creates a `DialogueStackFrame` from a dictionary.

        Args:
            data: The dictionary to create the `DialogueStackFrame` from.

        Returns:
            The created `DialogueStackFrame`.
        """
        suspended = data.get("suspended_agent_frames")
        if suspended is not None and not isinstance(suspended, list):
            suspended = None
        return UserFlowStackFrame(
            frame_id=data["frame_id"],
            flow_id=data["flow_id"],
            step_id=data["step_id"],
            frame_type=FlowStackFrameType.from_str(data.get("frame_type")),
            suspended_agent_frames=suspended,
        )


class AgentState(str, Enum):
    INTERRUPTED = "interrupted"
    WAITING_FOR_INPUT = "waiting_for_input"

    @staticmethod
    def from_str(state: Optional[str]) -> AgentState:
        if state == AgentState.WAITING_FOR_INPUT.value:
            return AgentState.WAITING_FOR_INPUT
        elif state == AgentState.INTERRUPTED.value:
            return AgentState.INTERRUPTED
        else:
            raise InvalidAgentState(state)


@dataclass
class AgentStackFrame(BaseFlowStackFrame):
    agent_id: str = ""
    state: AgentState = AgentState.WAITING_FOR_INPUT
    metadata: Optional[Dict[str, Any]] = None

    @classmethod
    def type(cls) -> str:
        """Returns the type of the frame."""
        return "agent"

    @staticmethod
    def from_dict(data: Dict[str, Any]) -> AgentStackFrame:
        """Creates a `AgentStackFrame` from a dictionary.

        Args:
            data: The dictionary to create the `AgentStackFrame` from.

        Returns:
            The created `AgentStackFrame`.
        """
        return AgentStackFrame(
            frame_id=data["frame_id"],
            flow_id=data["flow_id"],
            step_id=data["step_id"],
            agent_id=data["agent_id"],
            state=AgentState.from_str(data["state"]),
            metadata=data.get("metadata"),
        )



from typing import Any, Dict, List

from rasa_sdk import Action, Tracker
from rasa_sdk.events import SlotSet
from rasa_sdk.executor import CollectingDispatcher

from actions.db import get_account


class CheckBalance(Action):
    def name(self) -> str:
        return "check_balance"

    def run(
        self, dispatcher: CollectingDispatcher, tracker: Tracker, domain: Dict[str, Any]
    ) -> List[Dict[str, Any]]:
        account = get_account(tracker.sender_id)
        return [SlotSet("current_balance", account.funds)]



"""Configuration module for the prompt-to-bot service."""

import os
from typing import Any, Dict

from rasa.shared.utils.yaml import read_yaml

# OpenAI Configuration
OPENAI_MODEL = os.getenv("OPENAI_MODEL", "gpt-4.1-2025-04-14")
OPENAI_TEMPERATURE = float(os.getenv("OPENAI_TEMPERATURE", "0.7"))
OPENAI_VECTOR_STORE_ID = os.getenv(
    "OPENAI_VECTOR_STORE_ID", "vs_685123376e288191a005b6b144d3026f"
)
OPENAI_MAX_VECTOR_RESULTS = int(os.getenv("OPENAI_MAX_VECTOR_RESULTS", "10"))
OPENAI_TIMEOUT = int(os.getenv("OPENAI_TIMEOUT", "30"))
# OpenAI Token Pricing Configuration (per 1,000 tokens)
COPILOT_INPUT_TOKEN_PRICE = float(os.getenv("COPILOT_INPUT_TOKEN_PRICE", "0.002"))
COPILOT_OUTPUT_TOKEN_PRICE = float(os.getenv("COPILOT_OUTPUT_TOKEN_PRICE", "0.0005"))
COPILOT_CACHED_TOKEN_PRICE = float(os.getenv("COPILOT_CACHED_TOKEN_PRICE", "0.002"))

# Server Configuration
BUILDER_SERVER_HOST = os.getenv("SERVER_HOST", "0.0.0.0")
BUILDER_SERVER_PORT = int(os.getenv("SERVER_PORT", "5050"))
MAX_RETRIES = int(os.getenv("MAX_RETRIES", "5"))
MAX_LOG_ENTRIES = int(os.getenv("MAX_LOG_ENTRIES", "30"))
HELLO_RASA_PROJECT_ID = os.getenv("HELLO_RASA_PROJECT_ID")


# Copilot History Storage Configuration
# Relative path from project root to copilot database
COPILOT_DB_RELATIVE_PATH = ".rasa/copilot.db"
# Absolute path to copilot database (relative to current working directory)
# TODO: this shouldn't exist and should be removed. All data relevant to a project
# should be stored in the project folder. We can't do that at the moment, because
# we have a global llmservice singelton that is independent of the project folder.
# we should move away from that singleton (or at least extract the project specific
# data from the singleton to a separate service).
COPILOT_HISTORY_SQLITE_PATH = os.path.join(os.getcwd(), COPILOT_DB_RELATIVE_PATH)

# CORS Configuration
_cors_origins_env = os.getenv("CORS_ORIGINS", "*")
CORS_ORIGINS = _cors_origins_env.split(",") if _cors_origins_env != "*" else ["*"]

# Validation Configuration
VALIDATION_FAIL_ON_WARNINGS = (
    os.getenv("VALIDATION_FAIL_ON_WARNINGS", "false").lower() == "true"
)
VALIDATION_MAX_HISTORY = None  # Could be configured if needed

# Copilot Response Handler Configuration
COPILOT_CONTROLLED_RESPONSE_MAX_TOKENS = 20
COPILOT_HANDLER_ROLLING_BUFFER_SIZE = 20
COPILOT_ASSISTANT_TRACKER_MAX_TURNS = 10
COPILOT_DOCUMENTATION_SEARCH_QUERY_HISTORY_MESSAGES = 5

# Guardrail Configuration
ENABLE_GUARDRAILS_RAW = os.getenv("ENABLE_GUARDRAILS", "false")
ENABLE_GUARDRAILS = ENABLE_GUARDRAILS_RAW.strip().lower() == "true"

# TODO: Replace with Open Source guardrails implementation once it's ready
LAKERA_BASE_URL = os.getenv("LAKERA_BASE_URL", "https://api.lakera.ai/v2").rstrip("/")
LAKERA_ASSISTANT_HISTORY_GUARDRAIL_PROJECT_ID = "project-5442154580"
LAKERA_COPILOT_HISTORY_GUARDRAIL_PROJECT_ID = "project-4637497223"

# Guardrails blocking configuration
GUARDRAILS_ENABLE_BLOCKING_RAW = os.getenv("GUARDRAILS_ENABLE_BLOCKING", "false")
GUARDRAILS_ENABLE_BLOCKING = GUARDRAILS_ENABLE_BLOCKING_RAW.strip().lower() == "true"
GUARDRAILS_USER_MAX_STRIKES = int(os.getenv("GUARDRAILS_USER_MAX_STRIKES", "3"))
GUARDRAILS_PROJECT_MAX_STRIKES = int(os.getenv("GUARDRAILS_PROJECT_MAX_STRIKES", "5"))
GUARDRAILS_BLOCK_DURATION_SECONDS = int(
    os.getenv("GUARDRAILS_BLOCK_DURATION_SECONDS", "0")
)  # 0 means indefinite

# Auth0 Configuration
AUTH0_DOMAIN = "login.hello.rasa.ai"
AUTH0_CLIENT_ID = "Gq5RdRwp174OFIfTz6Re9TZUseYDXUYE"
AUTH0_ISSUER = f"https://{AUTH0_DOMAIN}/"
JWKS_URL = f"{AUTH0_ISSUER}.well-known/jwks.json"

# Inkeep Configuration
INKEEP_BASE_URL = os.getenv("INKEEP_BASE_URL", "https://api.inkeep.com/v1").rstrip("/")

# LLM Proxy Configuration
HELLO_LLM_PROXY_BASE_URL = os.getenv("HELLO_LLM_PROXY_BASE_URL")
RASA_PRO_LICENSE = os.getenv("RASA_PRO_LICENSE")

# Compute proxy-aware base URLs
if HELLO_LLM_PROXY_BASE_URL:
    _proxy = HELLO_LLM_PROXY_BASE_URL.rstrip("/")
    INKEEP_BASE_URL = f"{_proxy}/documentation"
    LAKERA_BASE_URL = f"{_proxy}/guardrails"


# Number of minutes after FIRST_USED when authentication becomes required
# across routes (except when auth is globally disabled on the server).
# Default kept small for development; adjust as needed.
AUTH_REQUIRED_AFTER_MINUTES = int(os.getenv("AUTH_REQUIRED_AFTER_MINUTES", "480"))

DEFAULT_BOT_BUILDER_EMAIL = "noreply@rasa.com"


def get_default_credentials() -> Dict[str, Any]:
    """Get default credentials configuration."""
    default_credentials_yaml = """
    studio_chat:
      user_message_evt: "user_message"
      bot_message_evt: "bot_message"
      session_persistence: true
    """
    return read_yaml(default_credentials_yaml)



from typing import Dict

from pydantic import BaseModel, Field

from rasa.builder.copilot.models import ResponseCategory


class ClassificationResult(BaseModel):
    prediction: ResponseCategory
    expected: ResponseCategory


class PerClassMetrics(BaseModel):
    """Metrics for a single response category."""

    precision: float = Field(ge=0.0, le=1.0, description="Precision score")
    recall: float = Field(ge=0.0, le=1.0, description="Recall score")
    f1: float = Field(ge=0.0, le=1.0, description="F1 score")

    support: int = Field(ge=0, description="Number of actual occurrences.")

    true_positives: int = Field(ge=0, description="Number of true positives.")
    false_positives: int = Field(ge=0, description="Number of false positives.")
    false_negatives: int = Field(ge=0, description="Number of false negatives.")


class OverallClassificationMetrics(BaseModel):
    """Overall evaluation metrics."""

    micro_precision: float = Field(
        ge=0.0, le=1.0, description="Micro-averaged Precision"
    )
    macro_precision: float = Field(
        ge=0.0, le=1.0, description="Macro-averaged Precision"
    )
    weighted_avg_precision: float = Field(
        ge=0.0, le=1.0, description="Weighted Precision"
    )

    micro_recall: float = Field(ge=0.0, le=1.0, description="Micro-averaged Recall")
    macro_recall: float = Field(ge=0.0, le=1.0, description="Macro-averaged Recall")
    weighted_avg_recall: float = Field(ge=0.0, le=1.0, description="Weighted Recall")

    micro_f1: float = Field(ge=0.0, le=1.0, description="Micro-averaged F1 score")
    macro_f1: float = Field(ge=0.0, le=1.0, description="Macro-averaged F1 score")
    weighted_avg_f1: float = Field(ge=0.0, le=1.0, description="Weighted F1 score")

    support: int = Field(ge=0, description="Total number of occurrences.")

    true_positives: int = Field(ge=0, description="Total number of true positives.")
    false_positives: int = Field(ge=0, description="Total number of false positives.")
    false_negatives: int = Field(ge=0, description="Total number of false negatives.")


class MetricsSummary(BaseModel):
    """Complete metrics summary with per-class and overall metrics."""

    per_class: Dict[ResponseCategory, PerClassMetrics] = Field(
        description="Per-class metrics"
    )
    overall: OverallClassificationMetrics = Field(description="Overall metrics")



from __future__ import annotations

import dataclasses
import json
import re
import sys
from dataclasses import dataclass
from enum import Enum
from functools import lru_cache
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Optional,
    Set,
    Text,
    Tuple,
    Type,
)

import structlog
from jinja2 import Template

import rasa.shared.utils.common
import rasa.shared.utils.io
from rasa.core.constants import DOMAIN_GROUND_TRUTH_METADATA_KEY
from rasa.core.policies.enterprise_search_policy import SEARCH_RESULTS_METADATA_KEY
from rasa.dialogue_understanding.patterns.clarify import FLOW_PATTERN_CLARIFICATION
from rasa.e2e_test.constants import (
    DEFAULT_ANSWER_RELEVANCE_PROMPT_TEMPLATE_FILE_NAME,
    DEFAULT_GROUNDEDNESS_PROMPT_TEMPLATE_FILE_NAME,
    KEY_GROUND_TRUTH,
    KEY_THRESHOLD,
    KEY_UTTER_NAME,
    KEY_UTTER_SOURCE,
    LLM_JUDGE_PROMPTS_MODULE,
)
from rasa.e2e_test.utils.generative_assertions import (
    ScoreInputs,
    _find_matching_generative_events,
    _parse_llm_output,
    _validate_parsed_llm_output,
    calculate_groundedness_score,
    calculate_relevance_score,
)
from rasa.shared.core.constants import DEFAULT_SLOT_NAMES
from rasa.shared.core.events import (
    ActionExecuted,
    BotUttered,
    DefinePrevUserUtteredFeaturization,
    DialogueStackUpdated,
    Event,
    FlowCancelled,
    FlowCompleted,
    FlowStarted,
    SlotSet,
)
from rasa.shared.exceptions import ProviderClientAPIException, RasaException
from rasa.shared.utils.llm import (
    llm_factory,
)
from rasa.utils.json_utils import SetEncoder

if TYPE_CHECKING:
    from rasa.e2e_test.e2e_config import LLMJudgeConfig


structlogger = structlog.get_logger()

DEFAULT_THRESHOLD = 0.5
OPERATOR_KEY = "operator"
FLOW_IDS_KEY = "flow_ids"


class AssertionType(Enum):
    FLOW_STARTED = "flow_started"
    FLOW_COMPLETED = "flow_completed"
    FLOW_CANCELLED = "flow_cancelled"
    PATTERN_CLARIFICATION_CONTAINS = "pattern_clarification_contains"
    ACTION_EXECUTED = "action_executed"
    SLOT_WAS_SET = "slot_was_set"
    SLOT_WAS_NOT_SET = "slot_was_not_set"
    BOT_UTTERED = "bot_uttered"
    BOT_DID_NOT_UTTER = "bot_did_not_utter"
    GENERATIVE_RESPONSE_IS_RELEVANT = "generative_response_is_relevant"
    GENERATIVE_RESPONSE_IS_GROUNDED = "generative_response_is_grounded"


class AssertionOperator(Enum):
    ANY = "any"
    ALL = "all"


@lru_cache(maxsize=1)
def _get_all_assertion_subclasses() -> Dict[str, Type[Assertion]]:
    return {
        sub_class.type(): sub_class
        for sub_class in rasa.shared.utils.common.all_subclasses(Assertion)
    }


class InvalidAssertionType(RasaException):
    """Raised if an assertion type is invalid."""

    def __init__(self, assertion_type: str) -> None:
        """Creates a `InvalidAssertionType`.

        Args:
            assertion_type: The invalid assertion type.
        """
        super().__init__(f"Invalid assertion type '{assertion_type}'.")


@dataclass
class Assertion:
    """Base class for storing assertions."""

    @classmethod
    def type(cls) -> str:
        """Returns the type of the assertion."""
        raise NotImplementedError

    @staticmethod
    def from_dict(assertion_dict: Dict[Text, Any]) -> Assertion:
        """Creates an assertion from a dictionary."""
        raise NotImplementedError

    def as_dict(self) -> Dict[str, Any]:
        """Return the `Assertion` as a dictionary.

        Returns:
            The `Assertion` as a dictionary.
        """
        data = dataclasses.asdict(self)
        data["type"] = self.type()
        if OPERATOR_KEY in data and data[OPERATOR_KEY] is not None:
            data[OPERATOR_KEY] = data[OPERATOR_KEY].value
        return data

    @staticmethod
    def create_typed_assertion(data: Dict[str, Any]) -> Assertion:
        """Creates a `Assertion` from a dictionary.

        Args:
            data: The dictionary to create the `Assertion` from.

        Returns:
            The created `Assertion`.
        """
        typ = next(iter(data.keys()))

        subclass_mapping = _get_all_assertion_subclasses()

        clazz = subclass_mapping.get(typ)

        if clazz is None:
            structlogger.warning("assertion.unknown_type", data=data)
            raise InvalidAssertionType(typ)

        try:
            return clazz.from_dict(data)
        except NotImplementedError:
            structlogger.warning("assertion.unknown_type", data=data)
            raise InvalidAssertionType(typ)

    def run(
        self,
        turn_events: List[Event],
        prior_events: List[Event],
        assertion_order_error_message: str = "",
        **kwargs: Any,
    ) -> Tuple[Optional[AssertionFailure], Optional[Event]]:
        """Run the assertion on the given events for that user turn.

        Args:
            turn_events: The events to run the assertion on.
            prior_events: All events prior to the current turn.
            assertion_order_error_message: The error message to append if the assertion
                order is enabled.
            kwargs: Additional keyword arguments.

        Returns:
            A tuple of the assertion failure and the matching event if the assertion
            passes, otherwise `None`.
        """
        raise NotImplementedError

    def generate_test_failure(
        self,
        error_message: str,
        assertion_order_error_message: str,
        prior_events: List[Event],
        turn_events: List[Event],
        line: Optional[int] = None,
    ) -> Tuple[AssertionFailure, None]:
        """Generate an assertion failure with the given error message."""
        error_message += assertion_order_error_message

        return self._generate_assertion_failure(
            error_message, prior_events, turn_events, line
        )

    def _generate_assertion_failure(
        self,
        error_message: str,
        prior_events: List[Event],
        turn_events: List[Event],
        line: Optional[int] = None,
    ) -> Tuple[AssertionFailure, None]:
        return AssertionFailure(
            assertion=self,
            error_message=error_message,
            actual_events_transcript=create_actual_events_transcript(
                prior_events, turn_events
            ),
            error_line=line,
        ), None


@dataclass
class FlowStartedAssertion(Assertion):
    """Class for storing the flow started assertion."""

    flow_id: Optional[str] = None
    line: Optional[int] = None
    operator: Optional[AssertionOperator] = None
    flow_ids: List[str] = dataclasses.field(default_factory=list)

    @classmethod
    def type(cls) -> str:
        return AssertionType.FLOW_STARTED.value

    @staticmethod
    def from_dict(assertion_dict: Dict[Text, Any]) -> FlowStartedAssertion:
        assertion_value = assertion_dict.get(AssertionType.FLOW_STARTED.value)
        operator = (
            assertion_value.get(OPERATOR_KEY, "").lower()
            if isinstance(assertion_value, dict)
            else None
        )
        flow_ids = (
            assertion_value.get(FLOW_IDS_KEY, [])
            if isinstance(assertion_value, dict)
            else []
        )
        return FlowStartedAssertion(
            flow_id=assertion_value if isinstance(assertion_value, str) else None,
            line=assertion_dict.lc.line + 1 if hasattr(assertion_dict, "lc") else None,
            operator=AssertionOperator(operator) if operator else None,
            flow_ids=flow_ids,
        )

    def run(
        self,
        turn_events: List[Event],
        prior_events: List[Event],
        assertion_order_error_message: str = "",
        **kwargs: Any,
    ) -> Tuple[Optional[AssertionFailure], Optional[Event]]:
        """Run the flow started assertion on the given events for that user turn."""
        if self.flow_id is not None:
            rasa.shared.utils.io.raise_deprecation_warning(
                f"'{AssertionType.FLOW_STARTED.value}' assertions defining "
                f"a single 'flow_id' value are "
                f"deprecated and will be removed in a future Rasa version. "
                f"Please use the '{FLOW_IDS_KEY}' field with an "
                f"'{OPERATOR_KEY}' instead.",
            )
            return self._run_single_flow_id_check(
                self.flow_id,
                turn_events,
                prior_events,
                assertion_order_error_message,
            )

        if self.operator == AssertionOperator.ALL:
            flow_id = next(
                iter(self.flow_ids)
            )  # there should be exactly one flow ID when using 'all' operator
            return self._run_single_flow_id_check(
                flow_id,
                turn_events,
                prior_events,
                assertion_order_error_message,
            )

        if self.operator == AssertionOperator.ANY:
            return self._run_any_flow_id_check(
                turn_events,
                prior_events,
                assertion_order_error_message,
            )

        raise RasaException(
            f"Invalid operator '{self.operator}' for 'flow_started' assertion. "
            "Supported operators are 'any' and 'all'."
        )

    def _run_single_flow_id_check(
        self,
        flow_id: str,
        turn_events: List[Event],
        prior_events: List[Event],
        assertion_order_error_message: str = "",
    ) -> Tuple[Optional[AssertionFailure], Optional[Event]]:
        """Run the deprecated flow_id check for backward compatibility."""
        try:
            matching_event = next(
                event
                for event in turn_events
                if isinstance(event, FlowStarted) and event.flow_id == flow_id
            )
        except StopIteration:
            error_message = f"Flow with id '{flow_id}' did not start."
            return self.generate_test_failure(
                error_message,
                assertion_order_error_message,
                prior_events,
                turn_events,
                self.line,
            )

        return None, matching_event

    def _run_any_flow_id_check(
        self,
        turn_events: List[Event],
        prior_events: List[Event],
        assertion_order_error_message: str = "",
    ) -> Tuple[Optional[AssertionFailure], Optional[Event]]:
        """Run the 'any' operator flow_id check."""
        for flow_id in self.flow_ids:
            try:
                matching_event = next(
                    event
                    for event in turn_events
                    if isinstance(event, FlowStarted) and event.flow_id == flow_id
                )
                return None, matching_event
            except StopIteration:
                continue

        error_message = (
            f"None of the flows with ids '{', '.join(self.flow_ids)}' started."
        )
        return self.generate_test_failure(
            error_message,
            assertion_order_error_message,
            prior_events,
            turn_events,
            self.line,
        )

    def __hash__(self) -> int:
        return hash(json.dumps(self.as_dict()))


@dataclass
class FlowCompletedAssertion(Assertion):
    """Class for storing the flow completed assertion."""

    flow_id: str
    flow_step_id: Optional[str] = None
    line: Optional[int] = None

    @classmethod
    def type(cls) -> str:
        return AssertionType.FLOW_COMPLETED.value

    @staticmethod
    def from_dict(assertion_dict: Dict[Text, Any]) -> FlowCompletedAssertion:
        line = assertion_dict.lc.line + 1 if hasattr(assertion_dict, "lc") else None
        assertion_dict = assertion_dict.get(AssertionType.FLOW_COMPLETED.value, {})

        return FlowCompletedAssertion(
            flow_id=assertion_dict.get("flow_id"),
            flow_step_id=assertion_dict.get("flow_step_id"),
            line=line,
        )

    def run(
        self,
        turn_events: List[Event],
        prior_events: List[Event],
        assertion_order_error_message: str = "",
        **kwargs: Any,
    ) -> Tuple[Optional[AssertionFailure], Optional[Event]]:
        """Run the flow completed assertion on the given events for that user turn."""
        try:
            matching_event = next(
                event
                for event in turn_events
                if isinstance(event, FlowCompleted) and event.flow_id == self.flow_id
            )
        except StopIteration:
            error_message = f"Flow with id '{self.flow_id}' did not complete."
            return self.generate_test_failure(
                error_message,
                assertion_order_error_message,
                prior_events,
                turn_events,
                self.line,
            )

        if (
            self.flow_step_id is not None
            and matching_event.step_id != self.flow_step_id
        ):
            error_message = (
                f"Flow with id '{self.flow_id}' did not complete "
                f"at expected step id '{self.flow_step_id}'. The actual "
                f"step id was '{matching_event.step_id}'."
            )
            return self.generate_test_failure(
                error_message,
                assertion_order_error_message,
                prior_events,
                turn_events,
                self.line,
            )

        return None, matching_event

    def __hash__(self) -> int:
        return hash(json.dumps(self.as_dict()))


@dataclass
class FlowCancelledAssertion(Assertion):
    """Class for storing the flow cancelled assertion."""

    flow_id: str
    flow_step_id: Optional[str] = None
    line: Optional[int] = None

    @classmethod
    def type(cls) -> str:
        return AssertionType.FLOW_CANCELLED.value

    @staticmethod
    def from_dict(assertion_dict: Dict[Text, Any]) -> FlowCancelledAssertion:
        line = assertion_dict.lc.line + 1 if hasattr(assertion_dict, "lc") else None
        assertion_dict = assertion_dict.get(AssertionType.FLOW_CANCELLED.value, {})

        return FlowCancelledAssertion(
            flow_id=assertion_dict.get("flow_id"),
            flow_step_id=assertion_dict.get("flow_step_id"),
            line=line,
        )

    def run(
        self,
        turn_events: List[Event],
        prior_events: List[Event],
        assertion_order_error_message: str = "",
        **kwargs: Any,
    ) -> Tuple[Optional[AssertionFailure], Optional[Event]]:
        """Run the flow cancelled assertion on the given events for that user turn."""
        try:
            matching_event = next(
                event
                for event in turn_events
                if isinstance(event, FlowCancelled) and event.flow_id == self.flow_id
            )
        except StopIteration:
            error_message = f"Flow with id '{self.flow_id}' was not cancelled."
            return self.generate_test_failure(
                error_message,
                assertion_order_error_message,
                prior_events,
                turn_events,
                self.line,
            )

        if (
            self.flow_step_id is not None
            and matching_event.step_id != self.flow_step_id
        ):
            error_message = (
                f"Flow with id '{self.flow_id}' was not cancelled "
                f"at expected step id '{self.flow_step_id}'. The actual "
                f"step id was '{matching_event.step_id}'."
            )
            return self.generate_test_failure(
                error_message,
                assertion_order_error_message,
                prior_events,
                turn_events,
                self.line,
            )

        return None, matching_event

    def __hash__(self) -> int:
        return hash(json.dumps(self.as_dict()))


@dataclass
class PatternClarificationContainsAssertion(Assertion):
    """Class for storing the pattern clarification contains assertion."""

    flow_names: Set[str]
    line: Optional[int] = None
    operator: Optional[AssertionOperator] = None
    flow_ids: List[str] = dataclasses.field(default_factory=list)

    @classmethod
    def type(cls) -> str:
        return AssertionType.PATTERN_CLARIFICATION_CONTAINS.value

    @staticmethod
    def from_dict(
        assertion_dict: Dict[Text, Any],
    ) -> PatternClarificationContainsAssertion:
        assertion_value = assertion_dict.get(
            AssertionType.PATTERN_CLARIFICATION_CONTAINS.value
        )

        operator = (
            assertion_value.get(OPERATOR_KEY, "").lower()
            if isinstance(assertion_value, dict)
            else None
        )

        flow_ids = (
            assertion_value.get("flow_ids", [])
            if isinstance(assertion_value, dict)
            else []
        )

        return PatternClarificationContainsAssertion(
            flow_names=set(assertion_value)
            if isinstance(assertion_value, list)
            else set(),
            line=assertion_dict.lc.line + 1 if hasattr(assertion_dict, "lc") else None,
            operator=AssertionOperator(operator) if operator else None,
            flow_ids=flow_ids,
        )

    def run(
        self,
        turn_events: List[Event],
        prior_events: List[Event],
        assertion_order_error_message: str = "",
        **kwargs: Any,
    ) -> Tuple[Optional[AssertionFailure], Optional[Event]]:
        """Run the flow completed assertion on the given events for that user turn."""
        try:
            matching_event = next(
                event
                for event in turn_events
                if isinstance(event, FlowStarted)
                and event.flow_id == FLOW_PATTERN_CLARIFICATION
            )
        except StopIteration:
            error_message = f"'{FLOW_PATTERN_CLARIFICATION}' pattern did not trigger."
            return self.generate_test_failure(
                error_message,
                assertion_order_error_message,
                prior_events,
                turn_events,
                self.line,
            )

        if self.operator is None and self.flow_names:
            return self._run_flow_names_check(
                matching_event,
                turn_events,
                prior_events,
                assertion_order_error_message,
            )

        actual_flow_ids = set(matching_event.metadata.get("clarification_ids", []))

        if self.operator == AssertionOperator.ALL:
            return self._run_all_flow_id_check(
                actual_flow_ids,
                matching_event,
                turn_events,
                prior_events,
                assertion_order_error_message,
            )

        if self.operator == AssertionOperator.ANY:
            return self._run_any_flow_id_check(
                actual_flow_ids,
                matching_event,
                turn_events,
                prior_events,
                assertion_order_error_message,
            )

        raise RasaException(
            f"Invalid operator '{self.operator}' for "
            f"'{FLOW_PATTERN_CLARIFICATION}' assertion. "
            "Supported operators are 'any' and 'all'."
        )

    def _run_flow_names_check(
        self,
        matching_event: Event,
        turn_events: List[Event],
        prior_events: List[Event],
        assertion_order_error_message: str = "",
    ) -> Tuple[Optional[AssertionFailure], Optional[Event]]:
        """Run the flow names check."""
        rasa.shared.utils.io.raise_deprecation_warning(
            f"'{AssertionType.PATTERN_CLARIFICATION_CONTAINS.value}' "
            f"assertions defining a list of flow names are deprecated "
            f"and will be removed in a future Rasa version. "
            f"Please use the '{FLOW_IDS_KEY}' field with an '{OPERATOR_KEY}' instead.",
        )

        actual_flow_names = set(matching_event.metadata.get("names", set()))
        if actual_flow_names != self.flow_names:
            error_message = (
                f"'{FLOW_PATTERN_CLARIFICATION}' pattern did not contain "
                f"the expected options. Expected options: {self.flow_names}. "
            )
            return self.generate_test_failure(
                error_message,
                assertion_order_error_message,
                prior_events,
                turn_events,
                self.line,
            )

        return None, matching_event

    def _run_all_flow_id_check(
        self,
        actual_flow_ids: Set[str],
        matching_event: Event,
        turn_events: List[Event],
        prior_events: List[Event],
        assertion_order_error_message: str = "",
    ) -> Tuple[Optional[AssertionFailure], Optional[Event]]:
        if actual_flow_ids != set(self.flow_ids):
            error_message = (
                f"'{FLOW_PATTERN_CLARIFICATION}' pattern did not contain all of "
                f"the expected options '{', '.join(self.flow_ids)}'. "
                f"Actual options: '{', '.join(actual_flow_ids)}'."
            )
            return self.generate_test_failure(
                error_message,
                assertion_order_error_message,
                prior_events,
                turn_events,
                self.line,
            )

        return None, matching_event

    def _run_any_flow_id_check(
        self,
        actual_flow_ids: Set[str],
        matching_event: Event,
        turn_events: List[Event],
        prior_events: List[Event],
        assertion_order_error_message: str = "",
    ) -> Tuple[Optional[AssertionFailure], Optional[Event]]:
        """Run the 'any' operator flow_id check."""
        for flow_id in self.flow_ids:
            if flow_id in actual_flow_ids:
                return None, matching_event

        error_message = (
            f"'{FLOW_PATTERN_CLARIFICATION}' pattern did not contain "
            f"any of the expected options '{', '.join(self.flow_ids)}'."
        )
        return self.generate_test_failure(
            error_message,
            assertion_order_error_message,
            prior_events,
            turn_events,
            self.line,
        )

    def __hash__(self) -> int:
        return hash(json.dumps(self.as_dict(), cls=SetEncoder))


@dataclass
class ActionExecutedAssertion(Assertion):
    """Class for storing the action executed assertion."""

    action_name: str
    line: Optional[int] = None

    @classmethod
    def type(cls) -> str:
        return AssertionType.ACTION_EXECUTED.value

    @staticmethod
    def from_dict(assertion_dict: Dict[Text, Any]) -> ActionExecutedAssertion:
        return ActionExecutedAssertion(
            action_name=assertion_dict.get(AssertionType.ACTION_EXECUTED.value),
            line=assertion_dict.lc.line + 1 if hasattr(assertion_dict, "lc") else None,
        )

    def run(
        self,
        turn_events: List[Event],
        prior_events: List[Event],
        assertion_order_error_message: str = "",
        **kwargs: Any,
    ) -> Tuple[Optional[AssertionFailure], Optional[Event]]:
        """Run the action executed assertion on the given events for that user turn."""
        step_index = kwargs.get("step_index")
        original_turn_events, turn_events = _get_turn_events_based_on_step_index(
            step_index, turn_events, prior_events
        )

        try:
            matching_event = next(
                event
                for event in turn_events
                if isinstance(event, ActionExecuted)
                and event.action_name == self.action_name
            )
        except StopIteration:
            error_message = f"Action '{self.action_name}' did not execute."
            return self.generate_test_failure(
                error_message,
                assertion_order_error_message,
                prior_events,
                original_turn_events,
                self.line,
            )

        return None, matching_event

    def __hash__(self) -> int:
        return hash(json.dumps(self.as_dict()))


@dataclass
class AssertedSlot:
    """Class for storing information asserted about slots."""

    name: str
    value: Any
    line: Optional[int] = None

    @staticmethod
    def from_dict(slot_dict: Dict[Text, Any]) -> AssertedSlot:
        return AssertedSlot(
            name=slot_dict.get("name"),
            value=slot_dict.get("value", "value key is undefined"),
            line=slot_dict.lc.line + 1 if hasattr(slot_dict, "lc") else None,
        )


@dataclass
class SlotWasSetAssertion(Assertion):
    """Class for storing the slot was set assertion."""

    slots: List[AssertedSlot]

    @classmethod
    def type(cls) -> str:
        return AssertionType.SLOT_WAS_SET.value

    @staticmethod
    def from_dict(assertion_dict: Dict[Text, Any]) -> SlotWasSetAssertion:
        return SlotWasSetAssertion(
            slots=[
                AssertedSlot.from_dict(slot)
                for slot in assertion_dict.get(AssertionType.SLOT_WAS_SET.value, [])
            ],
        )

    def run(
        self,
        turn_events: List[Event],
        prior_events: List[Event],
        assertion_order_error_message: str = "",
        **kwargs: Any,
    ) -> Tuple[Optional[AssertionFailure], Optional[Event]]:
        """Run the slot_was_set assertion on the given events for that user turn."""
        matching_event = None

        step_index = kwargs.get("step_index")
        original_turn_events, turn_events = _get_turn_events_based_on_step_index(
            step_index, turn_events, prior_events
        )

        for slot in self.slots:
            matching_events = [
                event
                for event in turn_events
                if isinstance(event, SlotSet) and event.key == slot.name
            ]
            if not matching_events:
                error_message = f"Slot '{slot.name}' was not set."
                return self.generate_test_failure(
                    error_message,
                    assertion_order_error_message,
                    prior_events,
                    original_turn_events,
                    slot.line,
                )

            if slot.value == "value key is undefined":
                matching_event = matching_events[0]
                structlogger.debug(
                    "slot_was_set_assertion.run",
                    last_event_seen=matching_event,
                    event_info="Slot value is not asserted and we have "
                    "multiple events for the same slot. "
                    "We will mark the first event as last event seen.",
                )
                continue

            try:
                matching_event = next(
                    event for event in matching_events if event.value == slot.value
                )
            except StopIteration:
                error_message = (
                    f"Slot '{slot.name}' was set to a different value "
                    f"'{matching_events[-1].value}' than the "
                    f"expected '{slot.value}' value."
                )
                return self.generate_test_failure(
                    error_message,
                    assertion_order_error_message,
                    prior_events,
                    original_turn_events,
                    slot.line,
                )

        return None, matching_event

    def __hash__(self) -> int:
        return hash(json.dumps(self.as_dict()))


@dataclass
class SlotWasNotSetAssertion(Assertion):
    """Class for storing the slot was not set assertion."""

    slots: List[AssertedSlot]

    @classmethod
    def type(cls) -> str:
        return AssertionType.SLOT_WAS_NOT_SET.value

    @staticmethod
    def from_dict(assertion_dict: Dict[Text, Any]) -> SlotWasNotSetAssertion:
        return SlotWasNotSetAssertion(
            slots=[
                AssertedSlot.from_dict(slot)
                for slot in assertion_dict.get(AssertionType.SLOT_WAS_NOT_SET.value, [])
            ]
        )

    def run(
        self,
        turn_events: List[Event],
        prior_events: List[Event],
        assertion_order_error_message: str = "",
        **kwargs: Any,
    ) -> Tuple[Optional[AssertionFailure], Optional[Event]]:
        """Run the slot_was_not_set assertion on the given events for that user turn."""
        matching_event = None

        step_index = kwargs.get("step_index")
        original_turn_events, turn_events = _get_turn_events_based_on_step_index(
            step_index, turn_events, prior_events
        )

        for slot in self.slots:
            matching_events = [
                event
                for event in turn_events
                if isinstance(event, SlotSet) and event.key == slot.name
            ]
            if not matching_events:
                continue

            # take the most recent event in the list of matching events
            # since that is the final value in the tracker for that user turn
            matching_event = matching_events[-1]

            if (
                slot.value == "value key is undefined"
                and matching_event.value is not None
            ):
                error_message = (
                    f"Slot '{slot.name}' was set to '{matching_event.value}' but "
                    f"it should not have been set."
                )
                return self.generate_test_failure(
                    error_message,
                    assertion_order_error_message,
                    prior_events,
                    turn_events,
                    slot.line,
                )

            if matching_event.value == slot.value:
                error_message = (
                    f"Slot '{slot.name}' was set to '{slot.value}' "
                    f"but it should not have been set."
                )
                return self.generate_test_failure(
                    error_message,
                    assertion_order_error_message,
                    prior_events,
                    original_turn_events,
                    slot.line,
                )

        return None, matching_event

    def __hash__(self) -> int:
        return hash(json.dumps(self.as_dict()))


@dataclass
class AssertedButton:
    """Class for storing information asserted about buttons."""

    title: str
    payload: Optional[str] = None

    @staticmethod
    def from_dict(button_dict: Dict[Text, Any]) -> AssertedButton:
        return AssertedButton(
            title=button_dict.get("title"),
            payload=button_dict.get("payload"),
        )


@dataclass
class BotUtteredAssertion(Assertion):
    """Class for storing the bot uttered assertion."""

    utter_name: Optional[str] = None
    text_matches: Optional[str] = None
    buttons: Optional[List[AssertedButton]] = None
    line: Optional[int] = None

    @classmethod
    def type(cls) -> str:
        return AssertionType.BOT_UTTERED.value

    @staticmethod
    def from_dict(assertion_dict: Dict[Text, Any]) -> BotUtteredAssertion:
        utter_name, text_matches, buttons = (
            BotUtteredAssertion._extract_assertion_properties(assertion_dict)
        )

        if BotUtteredAssertion._assertion_is_empty(utter_name, text_matches, buttons):
            raise RasaException(
                "A 'bot_uttered' assertion is empty, it should contain at least one "
                "of the allowed properties: 'utter_name', 'text_matches', 'buttons'."
            )

        return BotUtteredAssertion(
            utter_name=utter_name,
            text_matches=text_matches,
            buttons=buttons,
            line=assertion_dict.lc.line + 1 if hasattr(assertion_dict, "lc") else None,
        )

    @staticmethod
    def _extract_assertion_properties(
        assertion_dict: Dict[Text, Any],
    ) -> Tuple[Optional[str], Optional[str], List[AssertedButton]]:
        """Extracts the assertion properties from a dictionary."""
        assertion_dict = assertion_dict.get(AssertionType.BOT_UTTERED.value, {})
        utter_name = assertion_dict.get("utter_name")
        text_matches = assertion_dict.get("text_matches")
        buttons = [
            AssertedButton.from_dict(button)
            for button in assertion_dict.get("buttons", [])
        ]

        return utter_name, text_matches, buttons

    @staticmethod
    def _assertion_is_empty(
        utter_name: Optional[str],
        text_matches: Optional[str],
        buttons: List[AssertedButton],
    ) -> bool:
        """Validate if the bot uttered assertion is empty."""
        if not utter_name and not text_matches and not buttons:
            return True

        return False

    def run(
        self,
        turn_events: List[Event],
        prior_events: List[Event],
        assertion_order_error_message: str = "",
        **kwargs: Any,
    ) -> Tuple[Optional[AssertionFailure], Optional[Event]]:
        """Run the bot_uttered assertion on the given events for that user turn."""
        matching_event = None
        error_messages = []

        step_index = kwargs.get("step_index")
        original_turn_events, turn_events = _get_turn_events_based_on_step_index(
            step_index, turn_events, prior_events
        )

        if self.utter_name is not None:
            try:
                matching_event = next(
                    event
                    for event in turn_events
                    if isinstance(event, BotUttered)
                    and event.metadata.get("utter_action") == self.utter_name
                )
            except StopIteration:
                error_messages.append(
                    f"Bot did not utter '{self.utter_name}' response."
                )

        if self.text_matches is not None:
            pattern = re.compile(self.text_matches)
            try:
                matching_event = next(
                    event
                    for event in turn_events
                    if isinstance(event, BotUttered) and pattern.search(event.text)
                )
            except StopIteration:
                error_messages.append(
                    f"Bot did not utter any response which "
                    f"matches the provided text pattern "
                    f"'{self.text_matches}'."
                )

        if self.buttons:
            try:
                matching_event = next(
                    event
                    for event in turn_events
                    if isinstance(event, BotUttered) and self._buttons_match(event)
                )
            except StopIteration:
                error_messages.append(
                    "Bot did not utter any response with the expected buttons."
                )

        if error_messages:
            error_message = " ".join(error_messages)
            return self.generate_test_failure(
                error_message,
                assertion_order_error_message,
                prior_events,
                original_turn_events,
                self.line,
            )
        return None, matching_event

    def _buttons_match(self, event: BotUttered) -> bool:
        """Check if the bot response contains the expected buttons."""
        # a button is a dictionary with keys 'title' and 'payload'
        actual_buttons = event.data.get("buttons", [])
        if not actual_buttons:
            return False

        return all(
            self._button_matches(actual_button, expected_button)
            for actual_button, expected_button in zip(actual_buttons, self.buttons)
        )

    @staticmethod
    def _button_matches(
        actual_button: Dict[str, Any], expected_button: AssertedButton
    ) -> bool:
        """Check if the actual button matches the expected button."""
        return (
            actual_button.get("title") == expected_button.title
            and actual_button.get("payload") == expected_button.payload
        )

    def __hash__(self) -> int:
        return hash(json.dumps(self.as_dict()))


@dataclass
class BotDidNotUtterAssertion(Assertion):
    """Class for the 'bot_did_not_utter' assertion."""

    utter_name: Optional[str] = None
    text_matches: Optional[str] = None
    buttons: Optional[List[AssertedButton]] = None
    line: Optional[int] = None

    @classmethod
    def type(cls) -> str:
        return AssertionType.BOT_DID_NOT_UTTER.value

    @staticmethod
    def from_dict(assertion_dict: Dict[Text, Any]) -> BotDidNotUtterAssertion:
        """Creates a BotDidNotUtterAssertion from a dictionary."""
        assertion_dict = assertion_dict.get(AssertionType.BOT_DID_NOT_UTTER.value, {})
        utter_name = assertion_dict.get("utter_name")
        text_matches = assertion_dict.get("text_matches")
        buttons = [
            AssertedButton.from_dict(button)
            for button in assertion_dict.get("buttons", [])
        ]

        if not utter_name and not text_matches and not buttons:
            raise RasaException(
                "A 'bot_did_not_utter' assertion is empty. "
                "It should contain at least one of the allowed properties: "
                "'utter_name', 'text_matches', or 'buttons'."
            )

        return BotDidNotUtterAssertion(
            utter_name=utter_name,
            text_matches=text_matches,
            buttons=buttons,
            line=assertion_dict.lc.line + 1 if hasattr(assertion_dict, "lc") else None,
        )

    def run(
        self,
        turn_events: List[Event],
        prior_events: List[Event],
        assertion_order_error_message: str = "",
        **kwargs: Any,
    ) -> Tuple[Optional[AssertionFailure], Optional[Event]]:
        """Checks that the bot did not utter the specified messages or buttons."""
        step_index = kwargs.get("step_index")
        original_turn_events, turn_events = _get_turn_events_based_on_step_index(
            step_index, turn_events, prior_events
        )

        for event in turn_events:
            if isinstance(event, BotUttered):
                error_messages = []
                if self._utter_name_matches(event):
                    error_messages.append(
                        f"Bot uttered a forbidden utterance '{self.utter_name}'."
                    )
                if self._text_matches(event):
                    error_messages.append(
                        f"Bot uttered a forbidden message matching "
                        f"the pattern '{self.text_matches}'."
                    )
                if self._buttons_match(event):
                    error_messages.append(
                        "Bot uttered a forbidden response with specified buttons."
                    )

                if error_messages:
                    error_message = " ".join(error_messages)
                    return self.generate_test_failure(
                        error_message,
                        assertion_order_error_message,
                        prior_events,
                        original_turn_events,
                        self.line,
                    )
        return None, None

    def _utter_name_matches(self, event: BotUttered) -> bool:
        if self.utter_name is not None:
            if event.metadata.get("utter_action") == self.utter_name:
                return True
        return False

    def _text_matches(self, event: BotUttered) -> bool:
        if self.text_matches is not None:
            pattern = re.compile(self.text_matches)
            if pattern.search(event.text):
                return True
        return False

    def _buttons_match(self, event: BotUttered) -> bool:
        """Check if the bot response contains any of the forbidden buttons."""
        if self.buttons is None:
            return False

        actual_buttons = event.data.get("buttons", [])
        if not actual_buttons:
            return False

        for actual_button in actual_buttons:
            if any(
                self._is_forbidden_button(actual_button, forbidden_button)
                for forbidden_button in self.buttons
            ):
                return True
        return False

    @staticmethod
    def _is_forbidden_button(
        actual_button: Dict[str, Any], forbidden_button: AssertedButton
    ) -> bool:
        """Check if the button matches any of the forbidden buttons."""
        actual_title = actual_button.get("title")
        actual_payload = actual_button.get("payload")

        title_matches = forbidden_button.title == actual_title
        payload_matches = forbidden_button.payload == actual_payload
        if title_matches and payload_matches:
            return True
        return False

    def __hash__(self) -> int:
        """Hash method to ensure the assertion is hashable."""
        return hash(json.dumps(self.as_dict()))


@dataclass
class GenerativeResponseMixin(Assertion):
    """Mixin class for storing generative response assertions."""

    metric_adjective: str
    threshold: float = DEFAULT_THRESHOLD
    utter_name: Optional[str] = None
    utter_source: Optional[str] = None
    line: Optional[int] = None

    @classmethod
    def type(cls) -> str:
        return ""

    def as_dict(self) -> Dict[str, Any]:
        data = super().as_dict()
        data.pop("metric_adjective")
        return data

    def _render_prompt(self, matching_event: BotUttered) -> str:
        raise NotImplementedError

    def _get_processed_output(self, parsed_llm_output: Dict[str, Any]) -> List[Any]:
        raise NotImplementedError

    def _process_response(
        self, llm_response: str, bot_message: str
    ) -> List[Dict[str, Any]]:
        """Process the LLM response."""
        parsed_llm_output = _parse_llm_output(llm_response, bot_message)
        _validate_parsed_llm_output(parsed_llm_output, bot_message)

        processed_output = self._get_processed_output(parsed_llm_output)
        return processed_output

    def _run_llm_evaluation(
        self,
        matching_event: BotUttered,
        step_text: str,
        llm_judge_config: "LLMJudgeConfig",
        assertion_order_error_message: str,
        prior_events: List[Event],
        turn_events: List[Event],
    ) -> Tuple[Optional[AssertionFailure], Optional[Event]]:
        """Run the LLM evaluation on the given event."""
        bot_message = matching_event.text
        prompt = self._render_prompt(matching_event)
        llm_response = self._invoke_llm(llm_judge_config, prompt)

        try:
            processed_output = self._process_response(llm_response, bot_message)
        except RasaException as exc:
            structlogger.error(
                "e2e_test.generative_response_evaluation.error", error=exc
            )
            return self._generate_assertion_failure(
                str(exc), prior_events, turn_events, self.line
            )

        score_inputs = ScoreInputs(
            threshold=self.threshold,
            matching_event=matching_event,
            user_question=step_text,
            llm_judge_config=llm_judge_config,
        )
        score, error_justification = calculate_score(
            assertion_type=self.type(),
            processed_output=processed_output,
            score_inputs=score_inputs,
        )

        if score < self.threshold:
            error_message = (
                f"Generative response '{matching_event.text}' "
                f"given to the user input '{step_text}' "
                f"was not {self.metric_adjective}. "
                f"Expected score to be above '{self.threshold}' threshold, "
                f"but was '{round(score,2)}'. The LLM Judge model has justified its "
                f"score like so: {error_justification}."
            )
            return self.generate_test_failure(
                error_message,
                assertion_order_error_message,
                prior_events,
                turn_events,
                self.line,
            )

        return None, matching_event

    def _invoke_llm(self, llm_judge_config: LLMJudgeConfig, prompt: str) -> str:
        """Invoke the LLM to evaluate the generative response."""
        structlogger.debug(
            f"generative_response_is_{self.metric_adjective}_assertion.run_llm_evaluation",
        )

        llm = llm_factory(
            llm_judge_config.llm_config_as_dict,
            llm_judge_config.get_default_llm_config(),
        )

        try:
            llm_response = llm.completion(prompt)
            return llm_response.choices[0]
        except Exception as exc:
            structlogger.error(
                "e2e_test.generative_response_evaluation.llm.error", error=exc
            )
            raise ProviderClientAPIException(
                message="LLM call exception", original_exception=exc
            )

    def _run_assertion_with_utter_name(
        self,
        matching_events: List[BotUttered],
        step_text: str,
        llm_judge_config: "LLMJudgeConfig",
        assertion_order_error_message: str,
        prior_events: List[Event],
        turn_events: List[Event],
    ) -> Tuple[Optional[AssertionFailure], Optional[Event]]:
        """Assert metric for the given utter name."""
        try:
            matching_event = next(
                event
                for event in matching_events
                if event.metadata.get("utter_action") == self.utter_name
            )
        except StopIteration:
            error_message = f"Bot did not utter '{self.utter_name}' response."
            return self.generate_test_failure(
                error_message,
                assertion_order_error_message,
                prior_events,
                turn_events,
                self.line,
            )

        return self._run_llm_evaluation(
            matching_event,
            step_text,
            llm_judge_config,
            assertion_order_error_message,
            prior_events,
            turn_events,
        )

    def run(
        self,
        turn_events: List[Event],
        prior_events: List[Event],
        assertion_order_error_message: str = "",
        llm_judge_config: Optional["LLMJudgeConfig"] = None,
        step_text: Optional[str] = None,
        **kwargs: Any,
    ) -> Tuple[Optional[AssertionFailure], Optional[Event]]:
        """Run the LLM evaluation on the given events for that user turn."""
        matching_events: List[BotUttered] = _find_matching_generative_events(
            turn_events, self.utter_source
        )

        if not matching_events:
            error_message = (
                "No generative response issued by either the Enterprise Search Policy, "
                "IntentlessPolicy, Contextual Response Rephraser or a "
                "named custom action was found, "
                "but one was expected. Please define an utter_source in the assertion "
                "to specify which generative response to evaluate."
            )
            error_message += assertion_order_error_message

            return self._generate_assertion_failure(
                error_message, prior_events, turn_events, self.line
            )

        if self.utter_name is not None:
            return self._run_assertion_with_utter_name(
                matching_events,
                step_text,
                llm_judge_config,
                assertion_order_error_message,
                prior_events,
                turn_events,
            )

        if len(matching_events) > 1:
            structlogger.debug(
                f"generative_response_is_{self.metric_adjective}_assertion.run",
                event_info=f"Multiple generative responses found, "
                f"we will evaluate the first of the responses "
                f"'{matching_events[0].text}'.",
            )

        matching_event = matching_events[0]

        return self._run_llm_evaluation(
            matching_event,
            step_text,
            llm_judge_config,
            assertion_order_error_message,
            prior_events,
            turn_events,
        )


@dataclass
class GenerativeResponseIsRelevantAssertion(GenerativeResponseMixin):
    """Class for storing the generative response is relevant assertion."""

    @classmethod
    def type(cls) -> str:
        return AssertionType.GENERATIVE_RESPONSE_IS_RELEVANT.value

    def _render_prompt(self, matching_event: BotUttered) -> str:
        """Render the prompt."""
        inputs = _get_prompt_inputs(self.type(), matching_event)
        prompt_template = _get_default_prompt_template(
            DEFAULT_ANSWER_RELEVANCE_PROMPT_TEMPLATE_FILE_NAME
        )
        return Template(prompt_template).render(**inputs)

    @staticmethod
    def from_dict(
        assertion_dict: Dict[Text, Any],
    ) -> GenerativeResponseIsRelevantAssertion:
        assertion_dict = assertion_dict.get(
            AssertionType.GENERATIVE_RESPONSE_IS_RELEVANT.value, {}
        )

        return GenerativeResponseIsRelevantAssertion(
            threshold=assertion_dict.get(KEY_THRESHOLD, DEFAULT_THRESHOLD),
            utter_name=assertion_dict.get(KEY_UTTER_NAME),
            line=assertion_dict.lc.line + 1 if hasattr(assertion_dict, "lc") else None,
            metric_adjective="relevant",
            utter_source=assertion_dict.get(KEY_UTTER_SOURCE),
        )

    def __hash__(self) -> int:
        return hash(json.dumps(self.as_dict()))

    def _get_processed_output(self, parsed_llm_output: Dict[str, Any]) -> List[Any]:
        questions = parsed_llm_output.get("question_variations", [])
        if not questions:
            raise RasaException(
                "No question variations were extracted by the LLM Judge."
            )
        return questions


@dataclass
class GenerativeResponseIsGroundedAssertion(GenerativeResponseMixin):
    """Class for storing the generative response is grounded assertion."""

    ground_truth: Optional[str] = None

    @classmethod
    def type(cls) -> str:
        return AssertionType.GENERATIVE_RESPONSE_IS_GROUNDED.value

    def _render_prompt(self, matching_event: BotUttered) -> str:
        """Render the prompt."""
        inputs = _get_prompt_inputs(
            assertion_type=self.type(),
            matching_event=matching_event,
            ground_truth=self.ground_truth,
        )
        prompt_template = _get_default_prompt_template(
            DEFAULT_GROUNDEDNESS_PROMPT_TEMPLATE_FILE_NAME
        )
        return Template(prompt_template).render(**inputs)

    @staticmethod
    def from_dict(
        assertion_dict: Dict[Text, Any],
    ) -> GenerativeResponseIsGroundedAssertion:
        assertion_dict = assertion_dict.get(
            AssertionType.GENERATIVE_RESPONSE_IS_GROUNDED.value, {}
        )

        return GenerativeResponseIsGroundedAssertion(
            threshold=assertion_dict.get(KEY_THRESHOLD, DEFAULT_THRESHOLD),
            utter_name=assertion_dict.get(KEY_UTTER_NAME),
            ground_truth=assertion_dict.get(KEY_GROUND_TRUTH),
            line=assertion_dict.lc.line + 1 if hasattr(assertion_dict, "lc") else None,
            metric_adjective="grounded",
            utter_source=assertion_dict.get(KEY_UTTER_SOURCE),
        )

    def __hash__(self) -> int:
        return hash(json.dumps(self.as_dict()))

    def _get_processed_output(self, parsed_llm_output: Dict[str, Any]) -> List[Any]:
        """Process the LLM response."""
        statements = parsed_llm_output.get("statements", [])
        if not statements:
            raise RasaException(
                "No statements were extracted and scored by the LLM Judge. "
                "Please check the LLM Judge configuration"
            )

        return statements


@dataclass
class AssertionFailure:
    """Class for storing the assertion failure."""

    assertion: Assertion
    error_message: Text
    actual_events_transcript: List[Text]
    error_line: Optional[int] = None

    def as_dict(self) -> Dict[Text, Any]:
        """Returns the assertion failure as a dictionary."""
        return {
            "assertion": self.assertion.as_dict(),
            "error_message": self.error_message,
            "actual_events_transcript": self.actual_events_transcript,
        }


def create_actual_events_transcript(
    prior_events: List[Event], turn_events: List[Event]
) -> List[Text]:
    """Create the actual events transcript for the assertion failure."""
    all_events = prior_events + turn_events

    event_transcript = []

    for event in all_events:
        if isinstance(event, SlotSet) and event.key in DEFAULT_SLOT_NAMES:
            continue
        if isinstance(event, DefinePrevUserUtteredFeaturization):
            continue
        if isinstance(event, DialogueStackUpdated):
            continue

        event_transcript.append(repr(event))

    return event_transcript


def _get_turn_events_based_on_step_index(
    step_index: int, turn_events: List[Event], prior_events: List[Event]
) -> Tuple[List[Event], List[Event]]:
    """Get the turn events based on the step index.

    For the first step, we need to include the prior events as well
    in the same user turn. For the subsequent steps, we only need the
    events that follow the user uttered event on which the tracker
    was originally sliced by.

    Returns:
        List[Event]: The copy of turn_events
        List[Event]: The turn events based on the step index

    """
    original_turn_events = turn_events[:]
    if step_index == 0:
        return original_turn_events, prior_events + turn_events

    return original_turn_events, turn_events


def _get_default_prompt_template(default_prompt_template_file_name: str) -> str:
    # We cannot use importlib.resources with Python 3.9 because of an unfixed bug:
    # https://bugs.python.org/issue44137
    if sys.version_info < (3, 10):
        from importlib_resources import files

        default_prompt_template = (
            files(LLM_JUDGE_PROMPTS_MODULE)
            .joinpath(default_prompt_template_file_name)
            .read_text()
        )
    else:
        import importlib.resources

        default_prompt_template = importlib.resources.read_text(
            LLM_JUDGE_PROMPTS_MODULE,
            default_prompt_template_file_name,
        )

    return default_prompt_template


def _get_prompt_inputs(
    assertion_type: str,
    matching_event: BotUttered,
    ground_truth: Optional[str] = None,
) -> Dict[str, Any]:
    if assertion_type == AssertionType.GENERATIVE_RESPONSE_IS_RELEVANT.value:
        return {"num_variations": "3", "bot_message": matching_event.text}
    elif assertion_type == AssertionType.GENERATIVE_RESPONSE_IS_GROUNDED.value:
        ground_truth_event_metadata = matching_event.metadata.get(
            SEARCH_RESULTS_METADATA_KEY, ""
        ) or matching_event.metadata.get(DOMAIN_GROUND_TRUTH_METADATA_KEY, "")

        if isinstance(ground_truth_event_metadata, list):
            ground_truth_event_metadata = "\n".join(ground_truth_event_metadata)

        ground_truth = (
            ground_truth if ground_truth is not None else ground_truth_event_metadata
        )

        return {
            "bot_message": matching_event.text,
            "ground_truth": ground_truth,
        }
    else:
        raise ValueError(f"Invalid assertion type '{assertion_type}'")


def calculate_score(
    assertion_type: str, processed_output: List[Any], score_inputs: ScoreInputs
) -> Tuple[float, str]:
    """Calculate and return the score and justification."""
    if assertion_type == AssertionType.GENERATIVE_RESPONSE_IS_RELEVANT.value:
        return calculate_relevance_score(processed_output, score_inputs)
    elif assertion_type == AssertionType.GENERATIVE_RESPONSE_IS_GROUNDED.value:
        return calculate_groundedness_score(processed_output, score_inputs)
    else:
        raise ValueError(f"Invalid assertion type '{assertion_type}'")



from __future__ import annotations

import hmac
import json
import logging
import uuid
from base64 import b64encode
from functools import wraps
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Iterable,
    List,
    NoReturn,
    Optional,
    Text,
)

import jwt
from sanic import Blueprint, Sanic
from sanic.exceptions import Unauthorized
from sanic.request import Request

from rasa.cli import utils as cli_utils
from rasa.core.constants import BEARER_TOKEN_PREFIX
from rasa.shared.constants import DEFAULT_SENDER_ID, DOCS_BASE_URL
from rasa.shared.core.trackers import (
    DialogueStateTracker,
    EventVerbosity,
)
from rasa.shared.exceptions import RasaException

try:
    from urlparse import urljoin
except ImportError:
    from urllib.parse import urljoin

logger = logging.getLogger(__name__)


class UserMessage:
    """Represents an incoming message.

    Includes the channel the responses should be sent to.
    """

    def __init__(
        self,
        text: Optional[Text] = None,
        output_channel: Optional["OutputChannel"] = None,
        sender_id: Optional[Text] = None,
        parse_data: Optional[Dict[Text, Any]] = None,
        input_channel: Optional[Text] = None,
        message_id: Optional[Text] = None,
        metadata: Optional[Dict] = None,
        **kwargs: Any,
    ) -> None:
        """Creates a ``UserMessage`` object.

        Args:
            text: the message text content.
            output_channel: the output channel which should be used to send
                bot responses back to the user.
            sender_id: the message owner ID.
            parse_data: rasa data about the message.
            input_channel: the name of the channel which received this message.
            message_id: ID of the message.
            metadata: additional metadata for this message.
            **kwargs: additional arguments which will be included in the message.
        """
        self.text = text.strip() if text else text

        if message_id is not None:
            self.message_id = str(message_id)
        else:
            self.message_id = uuid.uuid4().hex

        if output_channel is not None:
            self.output_channel = output_channel
        else:
            self.output_channel = CollectingOutputChannel()

        if sender_id is not None:
            self.sender_id = str(sender_id)
        else:
            self.sender_id = DEFAULT_SENDER_ID

        self.input_channel = input_channel

        self.parse_data = parse_data
        self.metadata = metadata
        self.headers = kwargs.get("headers", None)

    def __repr__(self) -> Text:
        """Returns event as string for debugging."""
        return f"UserMessage(text: {self.text}, sender_id: {self.sender_id})"

    def __str__(self) -> Text:
        """Returns event as human-readable string."""
        return f"{self.__class__.__name__}({self.text})"


OnNewMessageType = Callable[[UserMessage], Awaitable[Any]]


def register(
    input_channels: List[InputChannel], app: Sanic, route: Optional[Text]
) -> None:
    """Registers input channel blueprints with Sanic."""

    async def handler(message: UserMessage) -> None:
        await app.ctx.agent.handle_message(message)

    for channel in input_channels:
        if route:
            p = urljoin(route, channel.url_prefix())
        else:
            p = None
        app.blueprint(channel.blueprint(handler), url_prefix=p)

    app.ctx.input_channels = input_channels


class InputChannel:
    """Input channel base class."""

    @classmethod
    def name(cls) -> Text:
        """Every input channel needs a name to identify it."""
        return cls.__name__

    @classmethod
    def from_credentials(cls, credentials: Optional[Dict[Text, Any]]) -> InputChannel:
        return cls()

    def url_prefix(self) -> Text:
        return self.name()

    def blueprint(self, on_new_message: OnNewMessageType) -> Blueprint:
        """Defines a Sanic blueprint.

        The blueprint will be attached to a running sanic server and handle
        incoming routes it registered for.
        """
        raise NotImplementedError("Component listener needs to provide blueprint.")

    @classmethod
    def raise_missing_credentials_exception(cls) -> NoReturn:
        raise RasaException(
            f"To use the {cls.name()} input channel, you need to "
            f"pass a credentials file using '--credentials'. "
            f"The argument should be a file path pointing to "
            f"a yml file containing the {cls.name()} authentication "
            f"information. Details in the docs: "
            f"{DOCS_BASE_URL}/messaging-and-voice-channels/"
        )

    def get_output_channel(self) -> Optional["OutputChannel"]:
        """Create ``OutputChannel`` based on information provided by the input channel.

        Implementing this function is not required. If this function returns a valid
        ``OutputChannel`` this can be used by Rasa to send bot responses to the user
        without the user initiating an interaction.

        Returns:
            ``OutputChannel`` instance or ``None`` in case creating an output channel
             only based on the information present in the ``InputChannel`` is not
             possible.
        """
        pass

    def get_metadata(self, request: Request) -> Optional[Dict[Text, Any]]:
        """Extracts additional information from the incoming request.

         Implementing this function is not required. However, it can be used to extract
         metadata from the request. The return value is passed on to the
         ``UserMessage`` object and stored in the conversation tracker.

        Args:
            request: incoming request with the message of the user

        Returns:
            Metadata which was extracted from the request.
        """
        pass


def decode_jwt(bearer_token: Text, jwt_key: Text, jwt_algorithm: Text) -> Dict:
    """Decodes a Bearer Token using the specific JWT key and algorithm.

    Args:
        bearer_token: Encoded Bearer token
        jwt_key: Public JWT key for decoding the Bearer token
        jwt_algorithm: JWT algorithm used for decoding the Bearer token

    Returns:
        `Dict` containing the decoded payload if successful or an exception
        if unsuccessful
    """
    authorization_header_value = bearer_token.replace(BEARER_TOKEN_PREFIX, "")
    return jwt.decode(authorization_header_value, jwt_key, algorithms=jwt_algorithm)


def decode_bearer_token(
    bearer_token: Text, jwt_key: Text, jwt_algorithm: Text
) -> Optional[Dict]:
    """Decodes a Bearer Token using the specific JWT key and algorithm.

    Args:
        bearer_token: Encoded Bearer token
        jwt_key: Public JWT key for decoding the Bearer token
        jwt_algorithm: JWT algorithm used for decoding the Bearer token

    Returns:
        `Dict` containing the decoded payload if successful or `None` if unsuccessful
    """
    # noinspection PyBroadException
    try:
        return decode_jwt(bearer_token, jwt_key, jwt_algorithm)
    except jwt.exceptions.InvalidSignatureError:
        logger.error("JWT public key invalid.")
    except Exception:
        logger.exception("Failed to decode bearer token.")

    return None


class OutputChannel:
    """Output channel base class.

    Provides sane implementation of the send methods
    for text only output channels.
    """

    def __init__(self) -> None:
        self.tracker_state: Optional[Dict[str, Any]] = None

    @classmethod
    def name(cls) -> Text:
        """Every output channel needs a name to identify it."""
        return cls.__name__

    def attach_tracker_state(self, tracker: DialogueStateTracker) -> None:
        """Attaches the current tracker state to the output channel."""
        self.tracker_state = tracker.current_state(EventVerbosity.AFTER_RESTART)

    async def send_response_chunk_start(
        self,
        recipient_id: Text,
        **kwargs: Any,
    ) -> None:
        """Indicates the start of a streaming response.

        Used for streaming generative AI responses. Default implementation does nothing.

        Args:
            recipient_id: The recipient ID.
            **kwargs: Additional arguments.
        """
        pass

    async def send_response_chunk(
        self,
        recipient_id: Text,
        chunk: Text,
        **kwargs: Any,
    ) -> None:
        """Send a chunk of response to the client.

        Used for streaming generative AI responses. Default implementation does nothing.

        Args:
            recipient_id: The recipient ID.
            chunk: The response chunk to send. Full response is built by concatenating
                all chunks sent.
            **kwargs: Additional arguments.

        Returns:
            None
        """
        pass

    async def send_response_chunk_end(
        self,
        recipient_id: Text,
        **kwargs: Any,
    ) -> None:
        """Indicates the end of a streaming response.

        Used for streaming generative AI responses. Default implementation does nothing.

        Args:
            recipient_id: The recipient ID.
            **kwargs: Additional arguments.
        """
        pass

    async def send_response(
        self,
        recipient_id: Text,
        message: Dict[Text, Any],
    ) -> None:
        """Send a message to the client."""
        if message.get("quick_replies"):
            await self.send_quick_replies(
                recipient_id,
                message.pop("text"),
                message.pop("quick_replies"),
                **message,
            )
        elif message.get("buttons"):
            await self.send_text_with_buttons(
                recipient_id, message.pop("text"), message.pop("buttons"), **message
            )
        elif message.get("text"):
            await self.send_text_message(recipient_id, message.pop("text"), **message)

        if message.get("custom"):
            await self.send_custom_json(recipient_id, message.pop("custom"), **message)

        # if there is an image we handle it separately as an attachment
        if message.get("image"):
            await self.send_image_url(recipient_id, message.pop("image"), **message)

        if message.get("attachment"):
            await self.send_attachment(
                recipient_id, message.pop("attachment"), **message
            )

        if message.get("elements"):
            await self.send_elements(recipient_id, message.pop("elements"), **message)

    async def send_text_message(
        self,
        recipient_id: Text,
        text: Text,
        **kwargs: Any,
    ) -> None:
        """Send a message through this channel."""
        raise NotImplementedError(
            "Output channel needs to implement a send message for simple texts."
        )

    async def send_image_url(
        self, recipient_id: Text, image: Text, **kwargs: Any
    ) -> None:
        """Sends an image. Default will just post the url as a string."""
        await self.send_text_message(recipient_id, f"Image: {image}")

    async def send_attachment(
        self, recipient_id: Text, attachment: Text, **kwargs: Any
    ) -> None:
        """Sends an attachment. Default will just post as a string."""
        await self.send_text_message(recipient_id, f"Attachment: {attachment}")

    async def send_text_with_buttons(
        self,
        recipient_id: Text,
        text: Text,
        buttons: List[Dict[Text, Any]],
        **kwargs: Any,
    ) -> None:
        """Sends buttons to the output.

        Default implementation will just post the buttons as a string.
        """
        await self.send_text_message(recipient_id, text)
        for idx, button in enumerate(buttons):
            button_msg = cli_utils.button_to_string(button, idx)
            await self.send_text_message(recipient_id, button_msg)

    async def send_text_with_buttons_concise(
        self,
        recipient_id: str,
        text: str,
        buttons: List[Dict[str, Any]],
        **kwargs: Any,
    ) -> None:
        """Sends buttons in a concise format, useful for voice channels."""
        if text.strip()[-1] not in {".", "!", "?", ":"}:
            text += "."
        text += " "
        for idx, button in enumerate(buttons):
            text += button["title"]
            if idx != len(buttons) - 1:
                text += ", "
        await self.send_text_message(recipient_id, text)

    async def send_quick_replies(
        self,
        recipient_id: Text,
        text: Text,
        quick_replies: List[Dict[Text, Any]],
        **kwargs: Any,
    ) -> None:
        """Sends quick replies to the output.

        Default implementation will just send as buttons.
        """
        await self.send_text_with_buttons(recipient_id, text, quick_replies)

    async def send_elements(
        self, recipient_id: Text, elements: Iterable[Dict[Text, Any]], **kwargs: Any
    ) -> None:
        """Sends elements to the output.

        Default implementation will just post the elements as a string.
        """
        for element in elements:
            element_msg = "{title} : {subtitle}".format(
                title=element.get("title", ""), subtitle=element.get("subtitle", "")
            )
            await self.send_text_with_buttons(
                recipient_id, element_msg, element.get("buttons", [])
            )

    async def send_custom_json(
        self, recipient_id: Text, json_message: Dict[Text, Any], **kwargs: Any
    ) -> None:
        """Sends json dict to the output channel.

        Default implementation will just post the json contents as a string.
        """
        await self.send_text_message(recipient_id, json.dumps(json_message))

    async def hangup(self, recipient_id: Text, **kwargs: Any) -> None:
        """Indicate that the conversation should be ended."""
        pass


class CollectingOutputChannel(OutputChannel):
    """Output channel that collects send messages in a list.

    (doesn't send them anywhere, just collects them).
    """

    def __init__(self) -> None:
        """Initialise list to collect messages."""
        self.messages: List[Dict[Text, Any]] = []

    @classmethod
    def name(cls) -> Text:
        """Name of the channel."""
        return "collector"

    @staticmethod
    def _message(
        recipient_id: Text,
        text: Optional[Text] = None,
        image: Optional[Text] = None,
        buttons: Optional[List[Dict[Text, Any]]] = None,
        attachment: Optional[Text] = None,
        custom: Optional[Dict[Text, Any]] = None,
    ) -> Dict:
        """Create a message object that will be stored."""
        obj = {
            "recipient_id": recipient_id,
            "text": text,
            "image": image,
            "buttons": buttons,
            "attachment": attachment,
            "custom": custom,
        }

        # filter out any values that are `None`
        return {k: v for k, v in obj.items() if v is not None}

    def latest_output(self) -> Optional[Dict[Text, Any]]:
        if self.messages:
            return self.messages[-1]
        else:
            return None

    async def _persist_message(self, message: Dict[Text, Any]) -> None:
        self.messages.append(message)

    async def send_text_message(
        self,
        recipient_id: Text,
        text: Text,
        **kwargs: Any,
    ) -> None:
        for message_part in text.strip().split("\n\n"):
            await self._persist_message(self._message(recipient_id, text=message_part))

    async def send_image_url(
        self, recipient_id: Text, image: Text, **kwargs: Any
    ) -> None:
        """Sends an image. Default will just post the url as a string."""
        await self._persist_message(self._message(recipient_id, image=image))

    async def send_attachment(
        self, recipient_id: Text, attachment: Text, **kwargs: Any
    ) -> None:
        """Sends an attachment. Default will just post as a string."""
        await self._persist_message(self._message(recipient_id, attachment=attachment))

    async def send_text_with_buttons(
        self,
        recipient_id: Text,
        text: Text,
        buttons: List[Dict[Text, Any]],
        **kwargs: Any,
    ) -> None:
        await self._persist_message(
            self._message(recipient_id, text=text, buttons=buttons)
        )

    async def send_custom_json(
        self, recipient_id: Text, json_message: Dict[Text, Any], **kwargs: Any
    ) -> None:
        await self._persist_message(self._message(recipient_id, custom=json_message))


BASIC_AUTH_SCHEME = "Basic"


def create_auth_requested_response_provider(
    realm: str,
) -> Callable[[Request, Any, Any], ResponseWithAuthRequested]:
    def _provider(
        request: Request, *args: Any, **kwargs: Any
    ) -> ResponseWithAuthRequested:
        return ResponseWithAuthRequested(scheme=BASIC_AUTH_SCHEME, realm=realm)

    return _provider


class ResponseWithAuthRequested(Unauthorized):
    """Custom exception to request authentication."""

    def __init__(self, scheme: str, realm: str) -> None:
        super().__init__(
            message="Authentication requested.", scheme=scheme, realm=realm
        )  # type: ignore[no-untyped-call]


def requires_basic_auth(
    username: Optional[Text],
    password: Optional[Text],
    auth_request_provider: Optional[Callable[[Request, Any, Any], Unauthorized]] = None,
) -> Callable:
    """Decorator to require basic auth for a route.

    Args:
        username: The username to check against.
        password: The password to check against.
        auth_request_provider: Optional function to provide a custom
            response when authentication is requested. This function should
            return an instance of `Unauthorized` with the necessary
            authentication headers.
    Returns:
        A decorator that checks for basic authentication.

    Raises:
        Unauthorized: If the authentication fails or if authentication is
            requested without necessary credentials (headers).
    """

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        async def wrapper(request: Request, *args: Any, **kwargs: Any) -> Any:
            if not username or not password:
                return await func(request, *args, **kwargs)

            auth_header = request.headers.get("Authorization")

            # Some systems will first send a request without an authorization header
            # to check if the endpoint is available. In this case, we need to
            # return am Unauthorized response with the necessary authorization header
            # to indicate that the endpoint requires authorization.
            if not auth_header and auth_request_provider:
                # if the request does not contain an authorization header,
                # we raise an exception to request authorization

                exception = auth_request_provider(request, *args, **kwargs)
                logger.debug(
                    f"Responding with {exception.status_code} and "
                    f"necessary auth headers {exception.headers}"
                )
                raise exception

            if not auth_header or not auth_header.startswith("Basic "):
                logger.error("Missing or invalid authorization header.")
                raise Unauthorized("Missing or invalid authorization header.")  # type: ignore[no-untyped-call]

            encoded = b64encode(f"{username}:{password}".encode()).decode()
            username_password_digest: str = auth_header[len(BASIC_AUTH_SCHEME) :]
            username_password_digest = username_password_digest.strip()

            if not hmac.compare_digest(username_password_digest, encoded):
                logger.error("Invalid username or password.")
                raise Unauthorized("Invalid username or password.")  # type: ignore[no-untyped-call]

            return await func(request, *args, **kwargs)

        return wrapper

    return decorator



import base64
import os
from typing import Optional

import structlog

from rasa.model_manager import config
from rasa.utils.io import subpath

structlogger = structlog.get_logger()


def write_encoded_data_to_file(encoded_data: bytes, file: str) -> None:
    """Write base64 encoded data to a file."""
    # create the directory if it does not exist of the parent directory
    os.makedirs(os.path.dirname(file), exist_ok=True)

    with open(file, "w") as f:
        decoded = base64.b64decode(encoded_data)
        text = decoded.decode("utf-8")
        f.write(text)


def logs_base_path() -> str:
    """Return the path to the logs' directory."""
    return subpath(config.SERVER_BASE_WORKING_DIRECTORY, "logs")


def ensure_base_directory_exists(directory: str) -> None:
    """Ensure that a files parent directory exists.

    Args:
        directory: The directory to check.
    """
    os.makedirs(os.path.dirname(directory), exist_ok=True)


def models_base_path() -> str:
    """Return the path to the models' directory."""
    return subpath(config.SERVER_BASE_WORKING_DIRECTORY, "models")


def logs_path(action_id: str) -> str:
    """Return the path to the log file for a given action id.

    Args:
        action_id: can either be a training_id or a deployment_id
    """
    return subpath(logs_base_path(), f"{action_id}.txt")


def get_logs_content(action_id: str) -> Optional[str]:
    """Return the content of the log file for a given action id."""
    try:
        with open(logs_path(action_id), "r") as file:
            return file.read()
    except FileNotFoundError:
        structlogger.debug("model_service.logs.not_found", action_id=action_id)
        return None
